/* lunar.h
 *
 *   LUNAR - Lightweight Underlay Network Ad hoc Routing
 *
 *   LUNAR logic header file
 *
 * Copyright (C) 2001-2005 <christian.tschudin@unibas.ch>
 *
 * This file is distributed under the terms of the GNU general Public
 * License (GPL), see the file COPYING
 */


// ----------------------------------------------------------------------
#ifndef LUNAR_H
#define LUNAR_H

#include "platform.h"

// all times are in microseconds

#define LUNAR_MIN_RING               1
#define LUNAR_MAX_RING               3

#define LUNAR_TIMEOUT_PER_RING   25000  // 25 msec RTT seems fair
#define LUNAR_MAX_RETRIES            3

#define LUNAR_FORCED_RRET      3000000  // route re-establishment time, in usec
#define LUNAR_RRET_JITTER       100000

#define LUNAR_IP_VERIFY_MAX          3  // if three probes fail, we claim the address
#define LUNAR_IP_PROBE_RETRY    500000  // interval before probing again

#define LUNAR_BRCAST_DELAY      200000  // grace period for path setup before blasting

#define LUNAR_DOTTED_DOMAIN_NAME  ".net.lunar"
#define LUNAR_MAX_HOSTNAME_LEN	XRP_MAX_NAME_LEN
#define LUNAR_DNS_TIMEOUT     (60*1000*1000)	// 60 sec
#define LUNAR_NAME_TIMEOUT    (25*1000*1000)    // 25 sec

#define LUNAR_DSDN_TIMEOUT    (20*1000*1000)    // 20 sec

// ----------------------------------------------------------------------
// implementation constants/limits (not related to the LUNAR protocol):

#define LUNAR_IP_SUBNET     ((192<<24)|(168<<16)|(42<<8))
#define LUNAR_IP_NAMESERVER (LUNAR_IP_SUBNET | 0xfe)
#define LUNAR_IP_NETMASK    (0xffffff00)
#define LUNAR_IP_BCAST      (LUNAR_IP_SUBNET | ~LUNAR_IP_NETMASK)


/* LUNAR IPv6 prefix is fd00:2001:620:1::/64, i.e. unique local
   IPv6 unicast addresses. These addresses are autoconfigured:
   The LUNAR module is sending RA + prefix to the IPv6 stack */
#define LUNAR_IPV6_PREFIX { { { 0xfd,0x00,0x20,0x01,0x06,0x20,0x00,0x01,0,0,0,0,0,0,0,0 } } }

// -------------------IMPLEMENTATION SPECIFIC--------------------------

struct lunartarget_s {
	union {
		uint8_t  eth[6];       // for netbox_getL2addr()
		hit_t hit;
		struct in_addr ipv4;
#ifdef LUNAR_HAS_IPV6
		struct in6_addr ipv6;
#endif
#ifdef LUNAR_HAS_DNS
		char name[LUNAR_MAX_HOSTNAME_LEN + 1];
#endif
	} id;
	char family;
};


struct namedb_s {
	struct lunartarget_s name;
	unsigned long expires;
	short peerndx;
};

// ----------------------------------------------------------------------

struct uc_pending_s {
	sapfsel_t caller;
	sapfsel_t backsel; // local selector for backw data
	struct lunartarget_s target;
	char is_bidir;
	int udp_port, trans_id;
	int ndx;
	char qnamefamily;
};

struct uc_transit_pending_s {
	sapfsel_t caller;
	sapfsel_t mybacksel;
	sapfsel_t replysel;
	sapfsel_t backsel;
	unsigned char replyeth[ETH_ALEN];
	unsigned char backeth[ETH_ALEN];
	char is_bidir;
};

struct bcast_pending_s {
	sapfsel_t caller;
	sapfsel_t fwd_sel;
	struct lunartarget_s target;
	int ndx;
};

struct bc_fanout_s {
	sapfsel_t local;
	sapfsel_t forward;
	unsigned char eth[ETH_ALEN];
};

#ifdef LUNAR_HAS_DNS
struct dns_pending_s {
	sapfsel_t caller;
	int qid;
	struct lunartarget_s name;
};
#endif

// ----------------------------------------------------------------------

struct sapfdest_s {
	int ifnr;
	struct sockaddr dst;
	sapfsel_t sel;
};

#define LUNAR_MAX_NAMES           128  // max number of names
#define LUNAR_MAX_PEERS            32  // max number of active IP peers
#define LUNAR_MAX_PATHS            32  // max number of active routes

#define AF_NAME   44
#define AF_HIT    45
#define AF_ETH    46

#define TARGET_IS_HIT(t)  ((t)->family == AF_HIT)
#define TARGET_IS_IPV4(t) ((t)->family == AF_INET)
#define TARGET_IS_IPV6(t) ((t)->family == AF_INET6)
#define TARGET_IS_NAME(t) ((t)->family == AF_NAME)

#ifdef LUNAR_HAS_IPV6
# define _TARGET_EQUAL6(a,b) \
	|| ( ((a)->family == AF_INET6) && \
	    !memcmp(&(a)->id.ipv6, &(b)->id.ipv6, sizeof((a)->id.ipv6)) )
# define _TARGET_BROADCAST6(a) \
	|| ((a)->family == AF_INET6 && (uint8_t)((a)->id.ipv6.s6_addr[0]) == 0xff)
#else
# define _TARGET_EQUAL6(a,b)
# define _TARGET_BROADCAST6(a)
#endif

#ifdef LUNAR_HAS_DNS
# define _TARGET_EQUALN(a,b) \
	|| ( ((a)->family == AF_NAME) && \
	     !strncmp((a)->id.name,(b)->id.name, LUNAR_MAX_HOSTNAME_LEN) )
#else
# define _TARGET_EQUALN(a,b)
#endif

#define TARGET_EQUAL(a,b) ( \
	((a)->family == (b)->family) && ( \
	   ((a)->family == AF_INET && ((a)->id.ipv4.s_addr == (b)->id.ipv4.s_addr)) \
	|| ((a)->family == AF_HIT && (!memcmp((a)->id.hit, (b)->id.hit, sizeof(hit_t)))) \
	_TARGET_EQUAL6(a,b) _TARGET_EQUALN(a,b) ))

# define TARGET_IS_BROADCAST(t) \
	(((t)->family == AF_INET && (t)->id.ipv4.s_addr == htonl(LUNAR_IP_BCAST)) \
	 _TARGET_BROADCAST6(t) )


#ifndef NS_PORT
struct path_s {
#else
class path_s {
    public:  
#endif
	sapfsel_t callback;	      // local handler to remove
	sapfsel_t fwdsel;	      // where to forward to
	sapfsel_t backsel;            // where to backward to
	short int used;               // 0=unused, 1...=resolving, -1=resolved
	unsigned char hops;	      // distance
	unsigned char flags;	      // see below for flag values
	netbox_timer_t timer;	      // for path garbage collection
};
#define PATH_IS_MAPPED   0x01  // == resolved
#define PATH_IS_BIDIR    0x02
#define PATH_BCTREE      0x04

#ifndef NS_PORT
struct peer_s {
#else
class peer_s {
    public:  
#endif

/*
	struct lunartarget_s addr;    // peer's "addr" (ip)
	struct lunartarget_s hit;     // peer's identity (host id tag)
*/
	struct path_s *path;
	netbox_timer_t timer;	      // for garbage collection
	unsigned char flags;	      // see below for flag values
	sapfsel_t upssel;             // to access ups from path_timeout 
};
// #define PEER_HAS_NAME    0x10
// #define PEER_HAS_ADDR    0x20
// #define PEER_HAS_HIT     0x40

#define PEER_IS_FAKE	 0x01
#define PEER_IS_VALID	 0x10
#define PEER_IS_BCAST	 0x20
#define PEER_IS_GW       0x40
#define PEER_DNS_PENDING 0x80

#endif // LUNAR_H

// eof
