/* xrp.c
 *
 *   SAPF kernel module
 *
 *   header file for parsing and formatting support for
 *   XRP (extensible resolution protocol)
 *
 * Copyright (C) 2003-2004 <christian.tschudin@unibas.ch>
 *
 * This file is distributed under the terms of the GNU general Public
 * License (GPL), see the file COPYING
 *
 */

#ifndef XRP_H
#define XRP_H


/*
 * XRP message format: (see LUNAR Internet Draft for more details)

4-Bytes header:
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|1|         XRP command         |      ttl      |    reserved   |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

plus multiple parameter objects of the form (inspired from RSVP/CASP):
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|0|     length (bytes)          |     class     |  class-type   |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|                      ... contents ...                         |
|                                                               |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

plus end-of-list object
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
|1|            0                |
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

*/

#define XRP_CMD_BIT		0x8000
#define XRP_CMD_EOM		(XRP_CMD_BIT | 0x0000)
#define XRP_CMD_LUNAR_RREQ	(XRP_CMD_BIT | 0x0001)
#define XRP_CMD_LUNAR_RREP	(XRP_CMD_BIT | 0x0002)
#define XRP_CMD_LUNAR_CONCAST	(XRP_CMD_BIT | 0x0003)
#ifdef LUNAR_HAS_GW
#define XRP_CMD_LUNAR_DP	(XRP_CMD_BIT | 0x0004)
#endif
#define XRP_CMD_LUNAR_DSDN	(XRP_CMD_BIT | 0x0005)

#define XRP_CLASS_LUNAR_RREQ_SERIES	1   // unique number
#define XRP_CLASS_LUNAR_RREQ_REPLYTO	2   // where to send back the reply
#define XRP_CLASS_LUNAR_RREQ_TARGET	3   // the thing to resolve
#define XRP_CLASS_LUNAR_RREQ_BACKPTR	4
#define XRP_CLASS_LUNAR_RREQ_SRC	5
#define XRP_CLASS_LUNAR_RREQ_FWDPTR	6

#define XRP_CLASS_LUNAR_RREP_FWDPTR	1
#define XRP_CLASS_LUNAR_RREP_SRC	2

#define XRP_CLASS_LUNAR_CONCAST_APP	1

#define XRP_CLASS_LUNAR_GW		1
#define XRP_CLASS_LUNAR_HELLO		2
#define XRP_CLASS_LUNAR_DSDN_REPLYTO	3
#define XRP_CLASS_LUNAR_DSDN_CONFLICT	4
#define XRP_CLASS_LUNAR_DSDN_RESOLVED	5

#define XRP_CTYPE_UNKNOWN       0
#define XRP_CTYPE_SELECTOR      1
#define XRP_CTYPE_IPv4          2
#define XRP_CTYPE_IPv6          3
#define XRP_CTYPE_SEL_ETH       4
#define XRP_CTYPE_SEL_IPv4_UDP  5
#define XRP_CTYPE_HIT128        6
#define XRP_CTYPE_NAME          7
#define XRP_CTYPE_ETH           8

#define XRP_CTYPE__COUNT	9 // update xrp_param_len in
                                  // src/lunar_init.c if changed

typedef unsigned char hit_t[16];
#define XRP_MAX_NAME_LEN       63 // +1 for \0

struct xrp_head_s {
	uint16_t cmd;
	uint8_t ttl;
	uint8_t reserved;
}__attribute__((packed));

// Parameters:

struct xrp_common_s {
	uint16_t len;
	uint8_t parclass;
	uint8_t partype;
}__attribute__((packed));

struct xrp_sel_s {
	uint16_t len;
        uint8_t parclass:8;
        uint8_t partype:8;
	sapfsel_t sel;
}__attribute__((packed));

struct xrp_sel_eth_s {
	uint16_t len;
	uint8_t parclass;
	uint8_t partype;
	sapfsel_t sel;
	unsigned char eth[6];
//	unsigned char pad[2];
}__attribute__((packed));

struct xrp_sel_udp_s {
	uint16_t len;
	uint8_t parclass;
	uint8_t partype;
	sapfsel_t sel;
	struct in_addr ip;
	uint16_t udp;
//	uint16_t pad;
}__attribute__((packed));

struct xrp_ipv4_s {
	uint16_t len;
	uint8_t parclass;
	uint8_t partype;
	struct in_addr ip;
}__attribute__((packed));


#ifdef LUNAR_HAS_IPV6
struct xrp_ipv6_s {
	uint16_t len;
	uint8_t parclass;
	uint8_t partype;
	struct in6_addr ip6;
}__attribute__((packed));
#endif

struct xrp_name_s {
	uint16_t len;
	uint8_t parclass;
	uint8_t partype;
	char name[XRP_MAX_NAME_LEN + 1];
}__attribute__((packed));


struct xrp_hit128_s { // host identity tag
	uint16_t len;
	uint8_t parclass;
	uint8_t partype;
	hit_t hit;
}__attribute__((packed));


typedef union {
	struct xrp_common_s param;
	struct xrp_sel_s sel;
	struct xrp_sel_eth_s seleth;
	struct xrp_sel_udp_s seludp;
	struct xrp_ipv4_s ipv4;
#ifdef LUNAR_HAS_IPV6
	struct xrp_ipv6_s ipv6;
#endif
	struct xrp_name_s name;
	struct xrp_hit128_s hit128;
} xrp_param_t;


#endif // XRP_H
