/* lunar.c
 *
 *   LUNAR - Lightweight Underlay Network Ad hoc Routing
 *
 *   Linux kernel module wrapper for the LUNAR logic
 *
 * Copyright (C) 2001-2005 <christian.tschudin@unibas.ch>
 *
 * This file is distributed under the terms of the GNU general Public
 * License (GPL), see the file COPYING
 *
 *   2005-07-15 CJ: added "yellow-pages" (/proc/net/lunar)
                    previous proc is now /proc/net/lunardb
 *   2005-06-29 CJ: klunar_proc_info now displays all our IPv6 addresses
 *   2005-05-01 update for new namedb_s struct
 *   2005-04-02 updated for new path_s struct and peer info
 *   2005-01-25 Christophe Jelger: IPV6 stuff
 *   2004-11-16 added REGISTER_NETDEV stuff
 *   2004-11-13 Per Svensson: Added support for kernel 2.6
 *   2004-01-09 moved generic init stuff into src/lunar_init.c,
 *              added mackill parameter parsing
 *   2004-01-05 the sources are now managed by a CVS
 *   2003-11-20 adjusted debuglevels of printk messages, init/fin of
 *              module now gets announced
 */

#include "../include/platform.h"
#include "../include/version.h"

#include <net/neighbour.h>
#ifndef CONFIG_ARM
# include <net/arp.h>
# ifdef LUNAR_HAS_IPV6
#   include <net/ndisc.h>
# endif
#else
extern struct neigh_table arp_tbl;
# include <endian.h>
#endif

#include "../include/sapf.h"
#include "../include/xrp.h"
#include "../include/netbox.h"
#include "../include/lunar.h"
#include "../include/sapf_decl.h"
#include "../include/xrp_decl.h"
#include "../include/netbox_decl.h"
#include "../include/lunar_decl.h"

#include "../src/sapf.c"
#include "sapf_lnx.c"

int conf_min_ring = LUNAR_MIN_RING;
int conf_max_ring = LUNAR_MAX_RING;
int conf_timeout_per_ring = LUNAR_TIMEOUT_PER_RING;
int conf_max_retries = LUNAR_MAX_RETRIES;
int conf_fri = LUNAR_FORCED_RRET; // forced rediscovery interval
int conf_fri_jitter = LUNAR_RRET_JITTER;
int conf_rreq_gap = 0;
int conf_data_gap = 0;

int xrp_param_len[XRP_CTYPE__COUNT];

#define MAX_MACKILL 3
int mackill_cnt;
unsigned char mackill_list[MAX_MACKILL][ETH_ALEN];

int mysapfhandle;

sapfsel_t bcast_selector;
char bc_buf[2000];
int bc_len;
sapfsel_t bc_sel;
sapfsel_t *bc_selp;

sapfsel_t ipv4deliver;

struct net_device *basedev;
struct net_device *lunardev;

int dgwmode;

char dsdname[128];
char dsdnparamscopy[128];  /* copy of dsdnparams but with , replaced with \0 */
int dsdnvector;

int gw_flag;	// set to 1 when node is a gateway
netbox_timer_t gwtimer;
netbox_timer_t nametimer;
netbox_timer_t dsdntimer;

struct namedb_s namedb[LUNAR_MAX_NAMES];
struct peer_s   peers[LUNAR_MAX_PEERS];
struct path_s   paths[LUNAR_MAX_PATHS];

#include "../src/lunar_init.c"
#include "netbox_lnx.c"

/* for /proc/lunar */

#define HAS_HIT 0x01
#define HAS_IP4 0x02
#define HAS_IP6 0x04
#define HAS_NAME 0x08
#define IS_UP 0x10
#define IS_SEARCH 0x20
#define IS_GW 0x40

// ----------------------------------------------------------------------

MODULE_AUTHOR("christian.tschudin@unibas.ch");
MODULE_DESCRIPTION("LUNAR kernel module (" LUNAR_VERSION "), 2002-2005");
MODULE_LICENSE("GPL");

char *basedevname;
char *forwarding;
char *mackill;
char *dsdnparams;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
MODULE_PARM(basedevname, "s");
MODULE_PARM(debuglevel, "i");
MODULE_PARM(sapfdebug, "i");
MODULE_PARM(forwarding, "s");
MODULE_PARM(mackill, "s");
MODULE_PARM(dsdnparams, "s");
#else
module_param(basedevname, charp, 0000);
module_param(debuglevel, int, 0);
module_param(sapfdebug, int, 0);
module_param(forwarding, charp, 0000);
module_param(mackill, charp, 0000);
module_param(dsdnparams, charp, 0000);
#endif

MODULE_PARM_DESC(basedevname, "name of base device (default: first device "
		          "with wireless extensions)");
MODULE_PARM_DESC(debuglevel, "debug level 0-10");
MODULE_PARM_DESC(sapfdebug, "sapf debug level 0-10");
MODULE_PARM_DESC(forwarding, "default gatewaying (no, yes)");
MODULE_PARM_DESC(mackill, "suppress packets from the given MAC address/es");
MODULE_PARM_DESC(dsdnparams, "dsdn parameters");

// ----------------------------------------------------------------------

static int klunar_netdev_event(struct notifier_block *this,
			       unsigned long event, void *ptr);
static struct notifier_block klunar_notifier = {
	notifier_call: klunar_netdev_event,
};

static int klunar_netdev_event(struct notifier_block *this,
                              unsigned long event, void *ptr)
{
        struct net_device *dev = (struct net_device *) ptr;

	if (!dev)
		return NOTIFY_DONE;

        DEBUG(10, printk("klunar notifier event %08lx %s\n", event, dev->name));

        switch (event) {
        case NETDEV_REGISTER:
        case NETDEV_UP:
		DEBUG(10, printk("notifier up %s\n", dev->name));
		if (basedev)
			break;
		if (basedevname) {
			struct net_device *d = dev_get_by_name(basedevname);
			if (!d)
				break;
			dev_put(d);
			if (d->ifindex != dev->ifindex)
				break;
			if (!strcmp(basedevname, dev->name)) {
				DEBUG(1, printk("klunar - new interface %s cannot be the lunar device\n", basedevname));
				break;
			}
		} else {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9)
			if (!dev->get_wireless_stats)
#else
			if (!dev->wireless_data)
#endif
				break;
		}
		DEBUG(1, printk("klunar - new (wireless) interface %s found\n",
				 dev->name));
		basedev = dev;
		netbox_addL2if(dev->ifindex, myhandle);
                break;
        case NETDEV_UNREGISTER:
        case NETDEV_DOWN:
		DEBUG(10, printk("notifier down %s\n", dev->name));
                if (dev == basedev) {
                        DEBUG(1, printk("klunar - (wireless) interface %s went away\n",
					 dev->name));
			basedev = 0;
			netbox_dropL2if(dev->ifindex);
                }
                break;

        default:
                break;
        };

        return NOTIFY_DONE;
}

// ----------------------------------------------------------------------

#include "../src/lunar_core.c"
#include "../src/lunar_dhcp.c"
#include "../src/lunar_ipv6.c"
#include "../src/lunar_dns.c"
#include "../src/lunar_gw.c"
#include "../src/xrp.c"

// ----------------------------------------------------------------------

static int parse_mackill(void)
{
	char *pa[MAX_MACKILL], *cp;
	int i, j; // , ia[ETH_ALEN];

	DEBUG(10, printk("mackill parse: %s\n", mackill));

	cp = mackill;
	while (cp && mackill_cnt < MAX_MACKILL) {
		pa[mackill_cnt] = strsep(&cp, ",");
		if (!pa[mackill_cnt])
			break;
		mackill_cnt++;
	}
	for (i = 0; i < mackill_cnt; i++) {
		// lnx kernel bug in 2.4.X: sscanf format "%x" does not work ....
		cp = pa[i];
		for (j = 0; j < ETH_ALEN; j++, cp++) {
			mackill_list[i][j] = 0;
			for (; isxdigit(*cp); cp++) {
				mackill_list[i][j] =
					(mackill_list[i][j] << 4) |
					(*cp <= '9' ? (*cp - '0') :
					           ((*cp & 0x07) + 9 ));
			}
			if (*cp && *cp != ':') break;
		}
		if ( j != ETH_ALEN) {
			DEBUG(0, printk("mackill: error in MAC addr %s\n", pa[i]));
			mackill_cnt--;
			return -1;
		}

		DEBUG(5, printk("mackill +%s\n",
				lunar_eth2str(mackill_list[i])));
	}
	return 0;
}

// ----------------------------------------------------------------------

static int klunar_open(struct net_device *dev)
{
	netif_start_queue(dev);
        return 0;
}


static void klunar_ipv4(struct sk_buff *skb, struct net_device *dev)
{
	sapfsel_t sel;
	sapfpkt pkt;
	struct iphdr *ip;
	int i;

	DEBUG(10, printk("  "));
	for (i = 0; i < 50; i++)
		DEBUG(10, printk(" %02x", skb->data[i]));
	DEBUG(10, printk("\n"));

	pkt.data = skb->data + dev->hard_header_len;
	pkt.len = skb->len - dev->hard_header_len;
	pkt.ifnr = -1;
	ip = (struct iphdr*) (pkt.data);

	if (pkt.len >= (sizeof(struct iphdr)+sizeof(struct udphdr)) &&
		ip->protocol == IPPROTO_UDP) {
		struct udphdr *up = (struct udphdr *) (ip+1);
#ifdef LUNAR_HAS_DHCP
		if (ntohs(up->source) == 68 &&  // DHCP clnt
		    ntohs(up->dest) == 67) {    // DHCP srvr
			lunar_do_dhcp(&pkt);
			return;
		}
#endif // LUNAR_HAS_DHCP

#ifdef LUNAR_HAS_DNS
		if (ntohs(up->dest) == 53) {    // DNS
			if (lunar_do_dns(&pkt))
				return;
		}
#endif // LUNAR_HAS_DNS
	}

	if (!memcmp(skb->data, "\xff\xff\xff\xff\xff\xff", ETH_ALEN)){ // broadcast
		if (bcast_selector)
			sapf_demux(mysapfhandle, bcast_selector, &pkt);
		else
			lunar_do_bc_setup(&bcast_selector, &pkt);
		return;
	} // else unicast

	sel = 0;
	memcpy(((char*)&sel) + sizeof(sel) - ETH_ALEN, skb->data, ETH_ALEN);
	sapf_demux(mysapfhandle, sel, &pkt);
}


#ifdef LUNAR_HAS_IPV6
static void klunar_ipv6(struct sk_buff *skb, struct net_device *dev)
{
	sapfsel_t sel;
	sapfpkt pkt;
	struct ipv6hdr *ip6h;
	struct icmp6hdr *icmp6h;
	struct nd_msg *ndh;
	struct lunartarget_s target;

	// int i;
	// DEBUG(10, printk("  "));
	// for (i = 0; i < skb->len; i++)
	//	DEBUG(10, printk(" %02x", skb->data[i]));
	// DEBUG(10, printk("\n"));

	pkt.data = skb->data ; // + dev->hard_header_len --> I want to check eth header
	pkt.len = skb->len; // - dev->hard_header_len
	pkt.ifnr = -1;

	ip6h = (struct ipv6hdr*)(pkt.data+ETH_HLEN);
	icmp6h = (struct icmp6hdr*)(pkt.data+ETH_HLEN+sizeof(struct ipv6hdr));
	ndh = (struct nd_msg*)(pkt.data+ETH_HLEN+sizeof(struct ipv6hdr));

	// Trying to optimize response time
	// I comment parse_v6_data()
	// and add minimal code here
	// parse_v6_data(&pkt);

	if(ip6h->nexthdr == NEXTHDR_ICMP) {
		switch(icmp6h->icmp6_type) {
		case NDISC_ROUTER_SOLICITATION: {
			struct in6_addr lunar_ipv6_prefix = LUNAR_IPV6_PREFIX;
			lunar_ipv6_send_RA(&lunar_ipv6_prefix, 0, 0, 0xff);
			break;
		}
		case NDISC_NEIGHBOUR_SOLICITATION:
			memcpy( &(target.id.ipv6), &ndh->target, 16 );
			target.family = AF_INET6;
			lunar_do_ndisc(&target, 0, 0, AF_INET6);
			break;
		case ICMPV6_ECHO_REQUEST:
		case ICMPV6_ECHO_REPLY:
			sel = 0;
			memcpy(((char*)&sel)+sizeof(sel)-ETH_ALEN, pkt.data, ETH_ALEN);
			pkt.len = pkt.len - ETH_HLEN;
			pkt.data = pkt.data + ETH_HLEN;
			sapf_demux(mysapfhandle, sel, &pkt);
			break;
		default:
			DEBUG(9, printk("Unsupported ICMP6 type %d\n", icmp6h->icmp6_type));
			break;
		}	
	}

	if( ip6h->nexthdr == NEXTHDR_TCP || ip6h->nexthdr == NEXTHDR_UDP ) {
		sel = 0;
		memcpy(((char*)&sel)+sizeof(sel)-ETH_ALEN, pkt.data, ETH_ALEN);
		pkt.len = pkt.len - ETH_HLEN;
		pkt.data = pkt.data + ETH_HLEN;
		sapf_demux(mysapfhandle, sel, &pkt);
	}
}
#endif // LUNAR_HAS_IPV6


static int klunar_start_xmit(struct sk_buff *skb, struct net_device *dev)
{
        struct net_device_stats *stats = (struct net_device_stats *) dev->priv;
	struct ethhdr *ep = (struct ethhdr *) skb->data;

        DEBUG(10, printk("klunar - start_xmit (%d bytes)\n", skb->len));
	if (stats) {
		stats->tx_packets++;
		stats->tx_bytes += skb->len;
	}

        switch (ntohs(ep->h_proto)) {
	case ETH_P_ARP: {
		struct lunartarget_s target;
		DEBUG(10, printk("klunar - arp detected\n"));
		memcpy( &(target.id.ipv4), skb->data + 38, 4 );
		target.family = AF_INET;
		lunar_do_ndisc(&target, 0, 0, AF_INET);
		break;
	}
	case ETH_P_IP:
		DEBUG(10, printk("klunar - IPv4 data\n"));
		klunar_ipv4(skb, dev);
		break;
#ifdef LUNAR_HAS_IPV6
	case ETH_P_IPV6:
		DEBUG(10, printk("klunar - IPv6 data\n"));
		klunar_ipv6(skb, dev);
		break;
#endif // LUNAR_HAS_IPV6
	default:
		break;
	}
	kfree_skb(skb);

        return 0;
}


static int klunar_stop(struct net_device *dev)
{
	if (basedev) {
		int i = basedev->ifindex;
		basedev = 0;
		netbox_dropL2if(i);
	}
        netif_stop_queue(dev);

        return 0;
}


static struct net_device_stats *klunar_get_stats(struct net_device *dev)
{
        return (struct net_device_stats *) dev->priv;
}


// ----------------------------------------------------------------------

int NS_CLASS klunar_proc_info(char *buf, char **start, off_t offset, int len)
{
	int written = 0, i, cnt;
	struct peer_s *p;
	struct lunartarget_s t;
	struct inet6_dev *in6dev = (struct inet6_dev*) lunardev->ip6_ptr;
	struct inet6_ifaddr *cur;

	for (p = peers, i = LUNAR_MAX_PEERS, cnt = 0; i > 0; p++, i--)
		if (p->flags)
			cnt++;

	written += sprintf(buf, "# lunar (%d table entries, %d %s)\n", LUNAR_MAX_PEERS, cnt, cnt == 1 ? "peer" : "peers");

#define adjustlen()	{ \
	if (written < offset) { offset -= written; written = 0; } \
	if (len <= written) \
		goto out; \
}

#ifdef LUNAR_HAS_DNS
	netbox_getL3addr(&t, lunardev, AF_NAME);
	written += sprintf(buf + written, "\nname=%s\n", t.id.name);
#endif
	netbox_getL3addr(&t, lunardev, AF_INET);
	written += sprintf(buf + written, "addr=%s\n",
			   lunar_ip2str(t.id.ipv4.s_addr));
#ifdef LUNAR_HAS_IPV6
	if (in6dev && in6dev->addr_list) {
		for (cur = in6dev->addr_list; cur; cur = cur->if_next) {
			if (!IN6_IS_ADDR_UNSPECIFIED(&cur->addr))
				written += sprintf(buf + written, "addr=%s\n", lunar_ip62str(&cur->addr));
		}
	}
#endif
	netbox_getL3addr(&t, lunardev, AF_HIT);
	written += sprintf(buf + written, "hit=%s\n",
			   lunar_hit2str(&t.id.hit));

	for (p = peers, i = 0; i < LUNAR_MAX_PEERS; i++, p++) {
		int j;
		if (!p->flags)
			continue;
		written += sprintf(buf + written, "\n");
		for (j = 0; j < LUNAR_MAX_NAMES; j++) {
			char *f, *cp;
			if (!namedb[j].name.family ||
			    namedb[j].peerndx != i)
				continue;
			switch(namedb[j].name.family) {
			case AF_HIT:
				f = "hit=%s\n";
				cp = lunar_hit2str(&namedb[j].name.id.hit);
				break;
			case AF_INET:
				f = "addr=%s\n";
				cp = lunar_ip2str(namedb[j].name.id.ipv4.s_addr);
				break;
#ifdef LUNAR_HAS_IPV6
			case AF_INET6:
				f = "addr=%s\n";
				cp = lunar_ip62str(&namedb[j].name.id.ipv6);
				break;
#endif
#ifdef LUNAR_HAS_DNS
			case AF_NAME:
				f = "name=%s\n";
				cp = namedb[j].name.id.name;
				break;
#endif
			default:
				f = cp = 0;
				break;
			}
			if (f)
				written += sprintf(buf + written, f, cp);
		}

		if (p->path) {
			struct sapfdest_s *d;
			if (p->path->used >= 0)
				written += sprintf(buf + written,
					   "search=%d\n",
					   p->path->used);
//			else
				written += sprintf(buf + written,
						   "pathlength=%d\n",
						   p->path->hops+1);

			d = (struct sapfdest_s*) SAPF_Instance
				sapf_getCallbackAux(mysapfhandle,
						    p->path->fwdsel);
			if (d)
				written += sprintf(buf + written,
				    "nexthop=%016Lx@%s\n", ntohsel(d->sel),
				    lunar_eth2str((unsigned char *)
						  d->dst.sa_data));
		}
		adjustlen();
	}

out:
	*start = buf + offset;
	written -= offset;
	if (written > len)
		return len;
	return (written < 0) ? 0 : written ;
}

int NS_CLASS klunar_proc_users(char *buf, char **start, off_t offset, int len)
{
	int written = 0, i, cnt;
	char flags = 0;
	struct peer_s *p;
	struct lunartarget_s t;
	
	for (p = peers, i = LUNAR_MAX_PEERS, cnt = 0; i > 0; p++, i--)
		if (p->flags)
			cnt++;

	written += sprintf(buf, "# lunar (%d table entries, %d %s)\n\n", LUNAR_MAX_PEERS, cnt, cnt == 1 ? "peer" : "peers");
	written += sprintf(buf + written, "FLAGS: N=name, H=hit, 4=IPv4, 6=IPv6, r=resolved | s=search, G=gateway\n\n");

#define adjustlen()	{ \
	if (written < offset) { offset -= written; written = 0; } \
	if (len <= written) \
		goto out; \
}

#ifdef LUNAR_HAS_DNS

	flags |= HAS_NAME;
#endif
	netbox_getL3addr(&t, lunardev, AF_INET);
	if (t.id.ipv4.s_addr)
		flags |= HAS_IP4;
#ifdef LUNAR_HAS_IPV6
	if (IamIPv6Ready())
		flags |= HAS_IP6;
#endif
	netbox_getL3addr(&t, lunardev, AF_HIT);
	flags |= HAS_HIT;

	if (gw_flag)
		flags |= IS_GW;
	
	written += print_flags(flags, buf + written);
	netbox_getL3addr(&t, lunardev, AF_NAME);
	written += sprintf(buf + written, "  %s", t.id.name);
	written += sprintf(buf + written, "\t--> Localhost");

	for (p = peers, i = 0; i < LUNAR_MAX_PEERS; i++, p++) {
		int j;
		struct namedb_s *t = 0;
		char *cp = 0;
		
		if (!p->flags)
			continue;
		written += sprintf(buf + written, "\n");
		flags = 0;
		for (j = 0; j < LUNAR_MAX_NAMES; j++) {
			if (!namedb[j].name.family ||
			    namedb[j].peerndx != i)
				continue;
			switch(namedb[j].name.family) {
			case AF_HIT:
				flags |= HAS_HIT;
				break;
			case AF_INET:
				flags |= HAS_IP4;
				break;
#ifdef LUNAR_HAS_IPV6
			case AF_INET6:
				flags |= HAS_IP6;
				break;
#endif
#ifdef LUNAR_HAS_DNS
			case AF_NAME:
				flags |= HAS_NAME;
				cp = namedb[j].name.id.name;
				t = &namedb[j];
				break;
#endif
			default:
				break;
			}
		}

		if (p->path) {
			if (p->path->used >= 0)
				flags |= IS_SEARCH;
//			else
				flags |= IS_UP;
		}
		
		if (p->flags & PEER_IS_GW)
			flags |= IS_GW;
		
		written += print_flags(flags, buf + written);
		
		if (cp)
			written += sprintf(buf + written, "  %s", cp);
		
		if ( p->timer.expires )	{
			if (t) {
				if ( p->timer.expires > t->expires )
					written += sprintf(buf + written, "\t--> Last heard %ld sec ago", 
					LUNAR_DNS_TIMEOUT/1000000-(p->timer.expires-jiffies)/HZ);
				else
					written += sprintf(buf + written, "\t--> Last heard %ld sec ago", 
					LUNAR_DNS_TIMEOUT/1000000-(t->expires-jiffies)/HZ);
			}
			else
				written += sprintf(buf + written, "\t--> Last heard %ld sec ago", 
					LUNAR_DNS_TIMEOUT/1000000-(p->timer.expires-jiffies)/HZ);
		}
		else
			written += sprintf(buf + written, "\t--> No lifetime");
		adjustlen();
	}
	written += sprintf(buf + written, "\n\n");
	adjustlen();

out:
	*start = buf + offset;
	written -= offset;
	if (written > len)
		return len;
	return (written < 0) ? 0 : written ;
}

int print_flags(char flags, char *buf)
{
	if (!buf )
		return 0;

	if (flags & HAS_NAME)
		sprintf(buf, "N");
	else
		sprintf(buf, ".");
	
	if (flags & HAS_HIT)
		sprintf(buf + 1, "H");
	else
		sprintf(buf + 1, ".");
	
	if (flags & HAS_IP4)
		sprintf(buf + 2, "4");
	else
		sprintf(buf + 2, ".");
	
	if (flags & HAS_IP6)
		sprintf(buf + 3, "6");
	else
		sprintf(buf + 3, ".");
	
	if (flags & IS_UP)
		sprintf(buf + 4, "r");
	else if (flags & IS_SEARCH)
		sprintf(buf + 4, "s");
	else
		sprintf(buf + 4, ".");
	
	if (flags & IS_GW)
		sprintf(buf + 5, "G");
	else
		sprintf(buf + 5, ".");
			
	return 6;
}


// ----------------------------------------------------------------------

static int __init init_klunar(void)
{
	struct net_device *dev = 0;
	int id;
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,17)
	char *sysname = utsname()->nodename;
#else
	char *sysname = system_utsname.nodename;
#endif	

	DEBUG(1, printk("klunar init module (%s)\n", sysname));
	if (sapf_init())
		return -1;
        proc_net_create("sapf", 0, sapf_proc_info);

	parse_mackill();

        lunardev = INIT_ETHERDEV(0, sizeof(struct net_device_stats));
	read_lock(&dev_base_lock);
	id = dev_alloc_name( lunardev, "lunar");
	read_unlock(&dev_base_lock);
	
        if (!lunardev) {
		DEBUG(10, printk("klunar_init: no lunardev\n"));
		sapf_close(mysapfhandle);
                return -ENOMEM;
	}
	
	SET_MODULE_OWNER(lunardev);

	if (basedevname) {
		dev = dev_get_by_name(basedevname);
		if (!dev) {
			DEBUG(1, printk("klunar - device %s not found\n", basedevname));
		} else {
			dev_put(dev);
			if (dev == lunardev) {
				DEBUG(1, printk("klunar - invalid base device %s\n",
				       basedevname));
				dev = 0;
			}
		}
		basedev = dev;
		if (dev)
			netbox_addL2if(dev->ifindex, mysapfhandle);
	} else {
		read_lock(&dev_base_lock);
		for (dev = dev_base; dev != NULL; dev = dev->next) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9)
			if (dev->get_wireless_stats)
#else
			if (dev->wireless_data)
#endif
				break;
		}
		if (dev) {
			basedev = dev;
			DEBUG(10, printk("klunar - wireless interface is %s\n",
				 dev->name));
			netbox_addL2if(dev->ifindex, mysapfhandle);
		}
		read_unlock(&dev_base_lock);
	}

        lunardev->open = &klunar_open;
        lunardev->hard_start_xmit = &klunar_start_xmit;
        lunardev->stop = &klunar_stop;
        lunardev->get_stats = &klunar_get_stats;

	lunardev->mtu -= sizeof(sapfsel_t);
        lunardev->tx_queue_len = 3;
        *(short*)(lunardev->dev_addr) = htons(0x00ff);
        get_random_bytes(lunardev->dev_addr + 2, 4);

	if (REGISTER_NETDEV(lunardev)) {
                DEBUG(10, printk("klunar_init: register_netdev failed\n"));	
		unregister_netdev(lunardev);
		goto out_free;
	}

	DEBUG(1, printk("klunar - lunar device is %s (%d)\n", lunardev->name, id));

	if (lunar_init()) {
		unregister_netdev(lunardev);
		goto out_free;
	}

        register_netdevice_notifier(&klunar_notifier);
        proc_net_create("lunardb", 0, klunar_proc_info);
	proc_net_create("lunar", 0, klunar_proc_users);

	return 0;
out_free:
	free_netdev(lunardev);
	lunardev = 0;
	return -1;
}


static void __exit exit_klunar(void)
{
	DEBUG(1, printk("klunar unloading module\n"));

        proc_net_remove("lunar");
	proc_net_remove("lunardb");
        unregister_netdevice_notifier(&klunar_notifier);
	unregister_netdev(lunardev);
	free_netdev(lunardev);
	lunardev = 0;

	lunar_fini();

        proc_net_remove("sapf");
	sapf_fini();
}


module_init(init_klunar);
module_exit(exit_klunar);

// eof
