/* knetbox.c
 *
 *   LUNAR - Lightweight Underlay Network Ad hoc Routing
 *
 *   an internal API for access to layer 2+3 service points
 *
 * Copyright (C) 2001-2005 <christian.tschudin@unibas.ch>
 *
 * This file is distributed under the terms of the GNU general Public
 * License (GPL), see the file COPYING
 *
 *   2005-07-06 CJ: modified neigh_{unmap;map} to avoid IPv6 crash 
 *   2005-06-29 CJ: changed HIT creation in netbox_init
 *   2005-05-30 CJ: added longest prefix matching in netbox_getL3addr
 *   2005-05-13 found a way to get nd_tbl's addr, without nd_tbl
 *   2005-01-26 Christophe Jelger: IPv6 stuff 
 *   2004-11-13 Per Svensson: Added support for kernel 2.6
 *   2004-01-09 added mackill filter in l2recv()
 *   2003-11-24 added getHIT128()
 *   2003-11-20 added netbox_hasDGW(), removed netbox_arpannounce()
 */

#include "../include/netbox.h"
#include "../include/netbox_decl.h"
#include <linux/rtnetlink.h>	// for IPv6 address flags

// ----------------------------------------------------------------------

static struct packet_type sapfpacket;
static int myhandle;
static hit_t hit;

#ifdef LUNAR_HAS_IPV6
static struct neigh_table *my_nd_tbl_ptr = 0;
#endif

// ----------------------------------------------------------------------

static netbox_timer_t *timerlist;
static struct timer_list netbox_timer;

static void timer_click(unsigned long dummy)
{
	if (timerlist) {
		netbox_timer_t *t = timerlist;
		timerlist = timerlist->next;
		if (timerlist) {
			netbox_timer.expires = timerlist->expires;
			add_timer(&netbox_timer);
		}
		t->fct(t->data);
	}
}


void netbox_newtimer(netbox_timer_t *t,
		     void (*fct)(void *), void *data, unsigned long delta_usec)
{
	netbox_timer_t **pp;
	unsigned long expires;
	
	delta_usec /= 1000;
	expires = jiffies + HZ * delta_usec / 1000;

	DEBUG(DBG_PROC, printk("newtimer %ld ms\n", delta_usec));
	if (!timerlist)
		pp = &timerlist;
	else {
		for (pp = &timerlist; ; pp = &((*pp)->next)) {
			if (!*pp || time_after((*pp)->expires, expires))
				break;
		}
	}
	t->next = *pp;
	t->expires = expires;
	t->fct = fct;
	t->data = data;
	*pp = t;

	if (t == timerlist) {
		if (t->next)
			del_timer(&netbox_timer);
		netbox_timer.expires = timerlist->expires;
		add_timer(&netbox_timer);
	}

	return;
}

void netbox_deltimer(netbox_timer_t *t)
{
	netbox_timer_t **pp;

	if (t == timerlist) {
		del_timer(&netbox_timer);
		timerlist = timerlist->next;
		if (timerlist) {
			netbox_timer.expires = timerlist->expires;
			add_timer(&netbox_timer);
		}
		return;
	}

	for (pp = &timerlist; *pp; pp = &((*pp)->next)) {
		if (*pp == t) {
			*pp = t->next;
			return;
		}
	}
}

// ----------------------------------------------------------------------

int netbox_init()
{
	/* 2005-06-29 - changed by Christophe Jelger
	 *
	 * I changed this function to use the MAC address
	 * of the real wireless interface in the HIT --> this increases
	 * the probability of having a unique HIT since EUI-48 are
	 * supposed to be unique
	 *
	 * HIT = systemname (1st 6 letters) + MAC address 
	         + timeofday (in sec since epoch)
	 *
	 * If systemname is too short (<6 chars), "replay" name until
	 * 6 chars are reached
	*/

	#define N 6
	
	struct sockaddr sa;
	struct timeval tv;
	uint32_t sec;
	int ifnr = basedev ? basedev->ifindex : 0;
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,17)
	char *sysname = utsname()->nodename;
#else
	char *sysname = system_utsname.nodename;
#endif	
	int len = strlen(sysname);
	int i,j;
	char *cp = hit;

        init_timer(&netbox_timer);
	netbox_timer.function = timer_click;

	do_gettimeofday(&tv);
	sec = htonl(tv.tv_sec);

	NETBOX_Instance netbox_getL2addr(ifnr, &sa);
	memcpy(hit+6, (unsigned char *)sa.sa_data, 6);
	memcpy(hit+12, (uint32_t *)(&sec), 4);

	/* + systemname or MAC if systemname is not set */

	if (len == 0)
		memcpy(hit, (unsigned char *)sa.sa_data, 6);
	else {
		i = N/len;
		j = N%len;
		
		while (i)
		{
			strncpy(cp, sysname, len);
			cp += len; i--;
		}
		
		if (j)
			strncpy(cp, sysname, j);
	}

	return 0;
}


void netbox_fini(void)
{
	if (timerlist)
		del_timer(&netbox_timer);
}

// ----------------------------------------------------------------------
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
int l2recv(struct sk_buff *skb,
		  struct net_device *indev,
		  struct packet_type *pt)
#else
int l2recv(struct sk_buff *skb,
		  struct net_device *indev,
		  struct packet_type *pt,
		  struct net_device *extradev)
#endif
{
	struct sapfpkt_s *p = (struct sapfpkt_s *) &skb->cb;
	sapfsel_t dest;
	int i;

	for (i =0; i < mackill_cnt; i++) {
		if (!memcmp(ETHSRC(skb),mackill_list[i], ETH_ALEN)) {
			DEBUG(DBG_LINK, printk("knetbox l2recv: dropping packet from %s\n", lunar_eth2str(ETHSRC(skb))));
			goto done;
		}
	}

	memset(p, 0, sizeof(*p));
	p->snr = 0;
	p->ifnr = indev->ifindex;
	p->data = skb->nh.raw + sizeof(sapfsel_t);
	p->len = skb->len - sizeof(sapfsel_t);

	memcpy(&dest, skb->nh.raw, sizeof(dest));

	DEBUG(DBG_LINK, printk("sapf packet, %d (%d) bytes via %s: sel=0x%016Lx, refcnt=%d\n",
		skb->len, skb->len - sizeof(sapfsel_t),
		indev ? indev->name : "no-dev", ntohsel(dest),
		indev->refcnt.counter));

	sapf_demux(*(int*)AF_PACKET_PRIV(pt), dest, p);

 done:
	kfree_skb(skb);
	return 0;
}


int netbox_addL2if(int ifnr, int handle)
{
	struct net_device *dev = dev_get_by_index(ifnr);

	DEBUG(DBG_WARN, printk("netbox addL2if ifnr=%d\n", ifnr));

	if (sapfpacket.func) {
		DEBUG(DBG_WARN, printk("  sapf already bound\n"));
		return -1;
	}
	if (dev) {
		DEBUG(DBG_WARN, printk("  dev=%s\n", dev->name));
		dev->tx_queue_len = 1;
	} else {
		DEBUG(DBG_WARN, printk("  no dev!!\n"));
	}

	myhandle = handle;

	//	MOD_INC_USE_COUNT;

	sapfpacket.type = htons(SAPF_ETHERTYPE);
	sapfpacket.dev = dev;
	sapfpacket.func = l2recv;

	AF_PACKET_PRIV(&sapfpacket) = &myhandle;
	dev_add_pack(&sapfpacket);

#ifdef LUNAR_HAS_IPV6
	// arrg, nd_tbl is not an exported symbol (of the IPv6 module)
	// anymore, so we have to do some data struct traversal ...
	if (!my_nd_tbl_ptr) {
		struct inet6_dev *in6dev = (struct inet6_dev*) dev->ip6_ptr;
		if (in6dev && in6dev->nd_parms)
			my_nd_tbl_ptr = in6dev->nd_parms->tbl;
//		printk("my_nd_tbl_ptr %p %p\n", my_nd_tbl_ptr, &nd_tbl);
	}
#endif

	return 0;
}


int netbox_dropL2if(int ifnr)
{
	if (!sapfpacket.func)
		return -1;

	DEBUG(DBG_WARN, printk("netbox dropL2if\n"));

	dev_remove_pack(&sapfpacket);
	neigh_ifdown(&arp_tbl, sapfpacket.dev);

	dev_put(sapfpacket.dev);
	sapfpacket.func = 0;

	return 0;
}


/* int netbox_getHIT128(hit_t *buf)
{
	memcpy(buf, hit, sizeof(hit));
	return 0; // success
}
*/

int netbox_getL2addr(int ifnr, struct sockaddr *sa)
{
	struct net_device *dev = dev_get_by_index(ifnr);

	if (!dev) {
		DEBUG(DBG_WARN, printk("knetbox getL2addr: wrong index %d\n", ifnr));
		return -1;
	}

	DEBUG(DBG_PROC, printk("getL2addr: %s\n", lunar_eth2str(dev->dev_addr)));
	memcpy(sa->sa_data, dev->dev_addr, ETH_ALEN);

	dev_put(dev);

	return 0;
}


int netbox_getL3addr(struct lunartarget_s *name,
		     struct net_device *dev, char family)
{
	switch (family) {
	case AF_HIT:
		memcpy(name->id.hit, hit, sizeof(hit));
		name->family = AF_HIT;
		return 0;
	case AF_INET:
	{
		struct in_device *indev = (struct in_device*) dev->ip_ptr;

		// in case something goes wrong
		// we initially set the addr to zero
		memset(name, 0, sizeof(*name));
		if (!indev || !indev->ifa_list)
			return -1;

		name->id.ipv4.s_addr = indev->ifa_list->ifa_local;
		name->family = AF_INET;
		DEBUG_TARGET(DBG_ALL, "local", name);
		return 0;
	}

#ifdef LUNAR_HAS_DNS
	case AF_NAME:
	{
		char *cp;
		strcpy(name->id.name, dsdname);
				
		/* remove uppercase */
		for ( cp=name->id.name; *cp != '\0'; cp++ ) {
			if ( *cp >= 'A' && *cp <= 'Z' )
				*cp += 32;
		}
		
		cp = strchr(name->id.name, '.');
		if (!cp)
			cp = name->id.name + strlen(name->id.name);
		if ((cp - name->id.name + strlen(LUNAR_DOTTED_DOMAIN_NAME))
						< LUNAR_MAX_HOSTNAME_LEN)
			strcpy(cp, LUNAR_DOTTED_DOMAIN_NAME);
		name->family = AF_NAME;
		return 0;
	}
#endif

#ifdef LUNAR_HAS_IPV6
	case AF_INET6:
	{
		struct inet6_dev *in6dev = (struct inet6_dev*) dev->ip6_ptr;
		struct inet6_ifaddr *cur;
		struct lunartarget_s target;
		__u16 scope = IFA_GLOBAL;
		int oldlen = -1, len = -1;

		memcpy(&target, name, sizeof(struct lunartarget_s));
		// return NULL address in case of a problem (+family for sanity checks)
		memset(&name->id.ipv6, 0, sizeof(struct in6_addr));
		name->family = AF_INET6;

		if (!in6dev || !in6dev->addr_list)
			return -1;

		if (IN6_IS_ADDR_LINKLOCAL(&target.id.ipv6))
			scope = IFA_LINK;
		else if (IN6_IS_ADDR_SITELOCAL(&target.id.ipv6))
			scope = IFA_SITE;

		DEBUG(DBG_ALL, printk("IPv6 target address (scope 0x%02x) is %s\n", scope, lunar_ip62str(&target.id.ipv6)));
		
		for (cur = in6dev->addr_list; cur; cur = cur->if_next) {
			DEBUG(DBG_ALL, printk("My IPv6 address is %s (prefix match %d)\n",
				lunar_ip62str(&cur->addr), prefix_match_len(&target.id.ipv6, &cur->addr)));
#ifdef LUNAR_HAS_GW
			if( cur->scope == IFA_GLOBAL && cur->flags&IFA_F_PERMANENT && 
			    !gw_flag && !IN6_IS_ADDR_MANETLOCAL(&cur->addr) ) {
				
				// one of our global address is permanent
				// i.e. it was set by root from user space
				// --> in LUNAR6 this means we are a gateway
				// but only if prefix is not MANET local
				
				DEBUG(DBG_ALL, printk("--> PERMANENT address (I'm a gateway)\n"));
				lunar_send_gwinfo(&cur->addr);
			}
#endif
			if (!memcmp(&target.id.ipv6,&(cur->addr),
						sizeof(struct in6_addr))) {
				// targeted address is found
				memcpy(&name->id.ipv6, &(cur->addr),
				       sizeof(struct in6_addr));
				break;
			}
			if (cur->scope == scope) {
				// Return our address with 
				// same scope as target
				// and longest prefix matching
				len = prefix_match_len(&target.id.ipv6, &cur->addr);
				if (len > oldlen) {
					oldlen = len;
					memcpy(&name->id.ipv6, &(cur->addr), sizeof(struct in6_addr));
				}
				// break; removed, since we now use longest prefix matching
			}
		}
		return 0;
	}
#endif
	default:
		return -1;
	}
}


int netbox_neigh_map(struct net_device *dev,
		     struct lunartarget_s *host, char *eth)
{
	struct neighbour *neigh = 0;

	// Always use __neigh_lookup_errno and not neigh_lookup
	// as the former creates an entry if none exists
	// NEEDED to setup paths in both directions

#ifdef LUNAR_HAS_DNS
	if( TARGET_IS_NAME(host) ) {
		DEBUG(DBG_WARN, printk("neigh_map called for %s ! (exiting)\n", host->id.name ));
		return -1;
	}
#endif
	DEBUG_TARGET(DBG_PROC, "neigh_map: from", host);
	if( TARGET_IS_IPV4(host) && host->id.ipv4.s_addr != 0 ) {
		neigh = __neigh_lookup_errno(&arp_tbl, (struct in_addr*)&host->id.ipv4, dev);
	}
#ifdef LUNAR_HAS_IPV6
	else if( TARGET_IS_IPV6(host) && !IN6_IS_ADDR_UNSPECIFIED(&host->id.ipv6)) {
		
		/* CJ: modified to handle special case where ipv6 module is
		 * loaded after lunar module
		 */
		
		if (!my_nd_tbl_ptr) {
			struct inet6_dev *in6dev = (struct inet6_dev*) dev->ip6_ptr;
			if (in6dev && in6dev->nd_parms)
				my_nd_tbl_ptr = in6dev->nd_parms->tbl;
		}
		
		if (my_nd_tbl_ptr)
			neigh = __neigh_lookup_errno(my_nd_tbl_ptr, &host->id.ipv6, dev);
		else
			return -1;
	}
#endif
	DEBUG(DBG_ALL, printk("neigh_map: to   %s\n", lunar_eth2str(eth)));

	//        err = PTR_ERR(neigh);
        if ( neigh && !IS_ERR(neigh)) {
		DEBUG(DBG_ALL, printk("neighb update\n"));
		neigh->parms->delay_probe_time = 0;

		// Is it the NEIGH_UPDATE_F_OVERRIDE and
		// NEIGH_UPDATE_F_ADMIN we actually use here?
                NEIGH_UPDATE(neigh, eth, NUD_REACHABLE, 1, 0);
                neigh_release(neigh);
        } else
		DEBUG(DBG_ALL, printk("no neighb!\n"));


	return 0;
}


int netbox_neigh_unmap(struct net_device *dev, struct lunartarget_s *host)
{
	struct neighbour *neigh = NULL;

	if (TARGET_IS_IPV4(host)) {
		neigh = neigh_lookup(&arp_tbl, &(host->id.ipv4.s_addr), dev);
	}

#ifdef LUNAR_HAS_IPV6
	else if (TARGET_IS_IPV6(host) && !IN6_IS_ADDR_UNSPECIFIED(&host->id.ipv6)) {
		
		/* CJ: modified to handle special case where ipv6 module is
		 * loaded after lunar module
		 */
		
		if (!my_nd_tbl_ptr) {
			struct inet6_dev *in6dev = (struct inet6_dev*) dev->ip6_ptr;
			if (in6dev && in6dev->nd_parms)
				my_nd_tbl_ptr = in6dev->nd_parms->tbl;
		}
		
		if (my_nd_tbl_ptr)
			neigh = neigh_lookup(my_nd_tbl_ptr, &host->id.ipv6, dev);
		else
			return -1;
	}
#endif
	else // if( TARGET_IS_NAME(host) )
		return -1;

        if (neigh) {
              	// Is it the NEIGH_UPDATE_F_OVERRIDE and
		// NEIGH_UPDATE_F_ADMIN we actually use here?
		DEBUG_TARGET(DBG_ALL, "neigh_unmap", host);
                NEIGH_UPDATE(neigh, 0, NUD_NONE, 1, 0);
                neigh_release(neigh);
        } else {
		DEBUG_TARGET(DBG_ALL, "neigh_unmap failed", host);
	}

	return 0;
}


int netbox_l2send(int ifnr, struct sockaddr *dest, sapfsel_t sel,
			 void *data, int len)
     // selector must be in network order
{
	struct net_device *dev;
	struct sk_buff *skb;
	char *cp;
	static struct sockaddr broadcast = {AF_UNSPEC, {0xff, 0xff, 0xff, 0xff, 0xff, 0xff}};

	if (!dest)
		dest = &broadcast;

	DEBUG(DBG_LINK, printk("netbox_l2send %s 0x%016Lx (%d bytes)\n",
	       lunar_eth2str(dest->sa_data), ntohsel(sel), len));

	dev = dev_get_by_index(ifnr);
	if (!dev) {
		DEBUG(DBG_WARN, printk("netbox_l2send: wrong ifnr\n"));
		return 0;
	}
	DEBUG(DBG_ALL, printk("  outgoing dev is %s\n", dev->name));

	skb = alloc_skb(dev->hard_header_len + 15 +
			sizeof(sel) + len, GFP_ATOMIC);
	if (!skb) {
		DEBUG(DBG_ALL, printk("  alloc_skb failed\n"));
		return 0;
	}

	skb_reserve(skb, (dev->hard_header_len+15)&~15);
	skb->nh.raw = skb->data;
	skb->protocol = htons(SAPF_ETHERTYPE);
	skb->dev = dev;

	cp = skb_put(skb, sizeof(sel) + len);
	memcpy(cp, &sel, sizeof(sel));
	memcpy(cp + sizeof(sel), data, len);

	if (dev->hard_header) {
		dev->hard_header(skb, dev, SAPF_ETHERTYPE,
				 dest->sa_data, 0, skb->len);
		dev_queue_xmit(skb);
	} else {
		DEBUG(DBG_WARN, printk("netbox_l2send: no hard header, what to do?\n"));
	}

	dev_put(dev);
	return len;
}

static uint16_t check( int count, uint16_t *data )
{
	uint32_t sum = 0;
	
	while( count > 1 ) {
	
		sum += *(data++);
		count -= 2;
	}
	
	if( count > 0 )
		sum += *(unsigned char *)data;
		
	while( sum>>16 )
		sum = (sum & 0xffff) + (sum>>16);
		
	return (~sum);
}

void NS_CLASS netbox_nat(sapfpkt *pkt, void *aux)
{
	struct iphdr *ip = (struct iphdr*) pkt->data;
	struct ipv6hdr *ip6 = (struct ipv6hdr*) pkt->data;
	struct in_device *indev = (struct in_device*) lunardev->ip_ptr;
	
	struct lunartarget_s src, *fake = 0;
	struct peer_s *cur = aux;
	
	struct udp_pseudo_hdr2 {
		uint32_t src, dst;
		unsigned char zero, proto;
		uint16_t len;
	};
	
	unsigned char pseudo_hdr[1500]; // for checksum calculation
	struct udp_pseudo_hdr2 *uph = (struct udp_pseudo_hdr2*) pseudo_hdr;
	int len;
	
	DEBUG(10, printk("lunar: natbox called with %d bytes\n", pkt->len));
	memset( pseudo_hdr, 0, 1500 );
	
	if (!aux || !indev)
	{
		DEBUG(10, printk("lunar: natbox peer pointer NULL ! (exit)\n"));
		return;
	}
	
	if ( !(cur->flags & PEER_IS_FAKE) )
	{
		DEBUG(10, printk("lunar: natbox peer not flaged as FAKE ! (exit)\n"));
		return;
	}
	
	if (ip->version == 4)
	{
		fake = namedb_getname(cur - peers, AF_INET, 0);
		
		if (!fake)
		{
			DEBUG(10, printk("lunar: natbox namedb entry not found ! (exit)\n"));
			return;
		}
		
		src.family = AF_INET;
		src.id.ipv4.s_addr = ip->saddr;
		
		/* NATify the IP header */
		
		ip->saddr = fake->id.ipv4.s_addr;
		ip->daddr = indev->ifa_list->ifa_local;
		ip->check = 0;
				
		/* NATify higher layer header */
		
		// build PSEUDO HEADER for checksum calculation
		
		uph->src = ip->saddr;
		uph->dst = ip->daddr;
		uph->proto = ip->protocol;
		len = ntohs(ip->tot_len) - sizeof(*ip);
					
		switch(ip->protocol)
		{
			case IPPROTO_UDP: {
				struct udphdr *udph = (struct udphdr*)(ip + 1);
				udph->check = 0;
				memcpy(uph+1, udph, len);
				uph->len = htons(len);
				udph->check = check(sizeof(*uph)+len, (u_int16_t*)uph);
				break;
			}
			case IPPROTO_TCP: {
				struct tcphdr *tcph = (struct tcphdr*)(ip + 1);
				tcph->check = 0;
				memcpy(uph+1, tcph, len);
				uph->len = htons(len);
				tcph->check = check(sizeof(*uph)+len, (u_int16_t*)uph);										break;
			}
			default:
				break;
		}
		
		/* do IP checksum */
		ip->check = ip_fast_csum((char *)ip, ip->ihl);
	}
	else if (ip6->version == 6)
	{
		struct lunartarget_s myaddr6;
		int len;
		
		fake = namedb_getname(cur - peers, AF_INET6, IPV6_ADDR_SCOPE_MANET);
		
		if (!fake)
		{
			DEBUG(10, printk("lunar: natbox namedb entry not found ! (exit)\n"));
			return;
		}
		
		src.family = AF_INET6;
		src.id.ipv6 = ip6->saddr;
		
		memset(&myaddr6, 0, sizeof(struct lunartarget_s));
		memset(&myaddr6, 0xfd, 1); // MANET scope
		NETBOX_Instance netbox_getL3addr(&myaddr6, lunardev, AF_INET6);
		
		/* NATify the IPv6 header */
		
		ip6->saddr = fake->id.ipv6;
		ip6->daddr = myaddr6.id.ipv6;
				
		/* NATify higher layer header */
		
		// build PSEUDO HEADER for checksum calculation
	
		memcpy( pseudo_hdr, &(ip6->saddr), 16 );
		memcpy( pseudo_hdr+16, &(ip6->daddr), 16 );
		memcpy( pseudo_hdr+32, &(ip6->payload_len), 2 );
		memcpy( pseudo_hdr+39, &(ip6->nexthdr), 1 );
		memcpy( pseudo_hdr+40, ip6 + 1, ntohs(ip6->payload_len) );
	
		len = ntohs(ip6->payload_len) + sizeof(struct ipv6hdr);
		
		switch(ip6->nexthdr)
		{
			case NEXTHDR_ICMP: {
				struct icmp6hdr *icmp6h = (struct icmp6hdr *)(pseudo_hdr + 40);
				icmp6h->icmp6_cksum = 0;
				icmp6h = (struct icmp6hdr *)(ip6 + 1);
				icmp6h->icmp6_cksum = check( len, (u_int16_t*)pseudo_hdr );
				break;
			}
			case NEXTHDR_TCP:{
				struct tcphdr *tcph = (struct tcphdr *)(pseudo_hdr + 40);
				tcph->check = 0;
				tcph = (struct tcphdr *) (ip6 + 1);
				tcph->check = check( len, (u_int16_t*)pseudo_hdr );
				break;
			}
			case NEXTHDR_UDP: {
				struct udphdr *udph = (struct udphdr *)(pseudo_hdr + 40);
				udph->check = 0;
				udph = (struct udphdr *) (ip6 + 1);
				udph->check = check( len, (u_int16_t*)pseudo_hdr );
				break;
			}
			default: {
				return;
				break;
			}
		}
	}
	
	/*
	DEBUG_TARGET(10, "lunar: NAT from", &src);
	DEBUG_TARGET(10, "lunar: NAT to", fake);*/
	
	L3deliver(pkt, aux);
}

int netbox_l3deliver(int ifnr, void *data, int len)
{
        struct net_device *dev = dev_get_by_index(ifnr);
        struct sk_buff *skb;

        if (!dev)
                return 0;

	DEBUG(DBG_SERVICE, printk("netbox_l3deliver %d Bytes\n", len));


	skb = alloc_skb(len + 14, GFP_ATOMIC);
	if (!skb) {
		dev_put(dev);
		return 0;
	}
	skb->dev = dev;

	skb->mac.raw = skb->data;
	skb_reserve(skb, 14);
	memcpy(skb_put(skb, len), data, len);
	memcpy(skb->mac.raw, dev->dev_addr, ETH_ALEN);
	memset(skb->mac.raw + ETH_ALEN, 0, ETH_ALEN);

	DEBUG(DBG_SERVICE, printk("netbox_l3deliver IP Version field is 0x%2x\n",
			 *(char*)(skb->mac.raw+2*ETH_ALEN+2)));

	if (*(char*)(skb->mac.raw + 2*ETH_ALEN +2) < 0x60) {
		// for IPv4, (char)version can be superior to 40
		// should be replaced with proper test using 4-bits bitfield

		*((short*)(skb->mac.raw + 2 * ETH_ALEN)) = htons(ETH_P_IP);
		skb->protocol = htons(ETH_P_IP);
	} else {
		*((short*)(skb->mac.raw + 2 * ETH_ALEN)) = htons(ETH_P_IPV6);
		skb->protocol = htons(ETH_P_IPV6);
	}
	netif_rx(skb);

	dev_put(dev);
	return 0;
}


/* int netbox_hasDGW(void);

   returns true when the host has a default GW entry in its
   routing table and should forward packets to the fixed
   internet.
*/
int netbox_hasDGW(void)
{
/*
	struct nlmsghdr nl;
	struct rtmsg rtm;
	struct kern_rta rta;
	struct rtentry r;
	int rc;
	struct fib_table *tb;

	memset(&r, 0, sizeof(r));
	r.rt_dst = ;
	r.rt_flags = ;
	r.rt_genmask = ;
	r.rt_gateway = ;
	rc = fib_convert_rtentry(SIOCDELRT, &nl, &rtm, &rta, &r);
	tb = fib_get_table(rtm.rtm_table);
*/
	return 0;
}

// eof
