/* lunar_core.c
 *
 *   LUNAR - Lightweight Underlay Network Ad hoc Routing
 *
 *   Lunar logic
 *
 * Copyright (C) 2001-2005 <christian.tschudin@unibas.ch>
 *
 * This file is distributed under the terms of the GNU general Public
 * License (GPL), see the file COPYING
 *
 *  2006-01-31 CJ: modified LUNARng version that implements
                   "private address maps"
		   --> added peer_genaddr() and modified a few things
		   to create peer addresses "on-demand"
 *  2005-07-15 CJ: added a lot of stuff to handle peer_timeout()
                   added "yellow-pages" (/proc/net/lunar)
 *  2005-07-06 CJ: added a lot of stuff to avoid doing silly things
                   when a node is not IPv6 enabled
 *  2005-07-04 CJ: added upssel in peer_s structure
 *                 --> used in path_timeout for expanding ring search
 *  2005-06-30 CJ: added basic peer merging in unicast_rrep_handler
                   modified lunar_rreq_handler_uc to extend rrep with IPv6 addr
 *  2005-06-29 CJ: skip doing RREQ for own IPv6 addresses (lunar_do_ndisc)
 *  2005-06-03 CJ: added scope parameter in namedb_get{name}
 *                 modified start_rreq to give namedb index
 *  2005-03-27 cft: send RREQ on activity with known # of hops,
 *             fixed ns2 init of path struct, misc DEBUG cleanup
 *  2005-03-xx cft: new path_s struct for handling names
 *  2005-01-25 Christophe Jelger: IPv6 stuff
 *  2004-08-xx Evgeny fixed a bug: RREQ for known hosts
 */


#include "../include/lunar.h"
#include "../include/lunar_decl.h"
#define NOT_PTR 2

// ----------------------------------------------------------------------

static char tmp[60];

#ifndef NO_DEBUG
void NS_CLASS debug_target(char *s, struct lunartarget_s *target)
{
	switch(target->family) {
	case AF_INET:
		printk("%s %s\n", s, lunar_ip2str(target->id.ipv4.s_addr));
		break;
#ifdef LUNAR_HAS_IPV6
	case AF_INET6:
		printk("%s %s\n", s, lunar_ip62str(&target->id.ipv6));
		break;
#endif
#ifdef LUNAR_HAS_DNS
	case AF_NAME:
		printk("%s '%s'\n", s, target->id.name);
		break;
#endif
	case AF_HIT:
		printk("%s %s\n", s, lunar_hit2str(&target->id.hit));
		break;
	default:
		printk("%s - unknown target family %d\n", s, target->family);
		break;
	}
}
#endif



char* NS_CLASS lunar_eth2str(char *data)
{
	sprintf(tmp, "%02x:%02x:%02x:%02x:%02x:%02x",
		 (unsigned char) data[0], (unsigned char) data[1],
		 (unsigned char) data[2], (unsigned char) data[3],
		 (unsigned char) data[4], (unsigned char) data[5]);

	return tmp;
}

char* NS_CLASS lunar_ip2str(int ip)
{
	ip = ntohl(ip);
	sprintf(tmp, "%d.%d.%d.%d",
		0x0ff & (ip >> 24), 0x0ff & (ip >> 16),
		0x0ff & (ip >> 8),  0x0ff & ip);

	return tmp;
}

char* NS_CLASS lunar_hit2str(hit_t *hit)
{
	int i;
	char *cp = tmp;

	for (i = 0; i < sizeof(*hit); i += 2) {
		if (i)
			*cp++ = '-';
		cp += sprintf(cp, "%02x%02x", (*hit)[i], (*hit)[i+1]);
	}
	*cp = '\0';
	return tmp;
}


#ifdef LUNAR_HAS_IPV6
char* NS_CLASS lunar_ip62str(struct in6_addr *ip6)
{
	int i;
	char *cp = tmp;
	
	for (i = 0; i < sizeof(struct in6_addr)/2; i++) {
		if (i)
			*cp++ = ':';
		if (ip6->s6_addr16[i] == 0)
			cp += sprintf(cp, "0");
		else if (ip6->s6_addr[2*i] == 0)
			cp += sprintf(cp, "%x", ip6->s6_addr[2*i+1]);
		else
			cp += sprintf(cp, "%x%02x",
				      ip6->s6_addr[2*i],
				      ip6->s6_addr[2*i+1]);
	}
	*cp = '\0';
	return tmp;
}
#endif // LUNAR_HAS_IPV6

// ----------------------------------------------------------------------

void NS_CLASS lunar_bcast_fanout(sapfpkt *p, void *aux)
{
	int ifnr = basedev ? basedev->ifindex : 0;
	struct bc_fanout_s *bf = (struct bc_fanout_s *)aux;

	if (bf->local)
		sapf_demux(mysapfhandle, bf->local, p);

	if (bf->forward)
		NETBOX_Instance netbox_l2send(ifnr,
//				bf->eth,
				0, bf->forward,
				p->data, p->len);
};


void NS_CLASS schedule_rreq(void *data, int len, unsigned long usec)
{
	int ifnr = basedev ? basedev->ifindex : 0;

	NETBOX_Instance netbox_l2send(ifnr, 0,
				      htonsel(SAPF_SELNET_LUNAR_RREQ),
				      data, len);
}

// ----------------------------------------------------------------------


int NS_CLASS namedb_get(int peerndx, char targetfamily, char scope)
{
	int i;

	for (i = 0; i < LUNAR_MAX_NAMES; i++) {
		if (namedb[i].name.family == targetfamily &&
		    namedb[i].peerndx == peerndx)
#ifdef LUNAR_HAS_IPV6
		{
			if (!scope)
				return i;
			else if ( TARGET_IS_IPV6(&namedb[i].name) &&
			    (scope == addr_scope(&namedb[i].name.id.ipv6)) )
				return i;
		}	
#else
		return i;
#endif
		
	}
	return -1;
}

int NS_CLASS namedb_getallipnames(int peerndx, int ndx, struct lunartarget_s *addr)
{
	int i;
	
	for (i = ndx; i < LUNAR_MAX_NAMES; i++) {
		if( namedb[i].peerndx == peerndx &&
		  ( TARGET_IS_IPV4(&namedb[i].name) 
#ifdef LUNAR_HAS_IPV6
		 || TARGET_IS_IPV6(&namedb[i].name)
#endif
		   ))
		{
			memcpy(addr, &namedb[i].name, sizeof(struct lunartarget_s));
			return i+1;
		}
	}
	addr = 0;
	return -1;
}

struct lunartarget_s* NS_CLASS namedb_getname(int peerndx, char targetfamily, char scope)
{
	int i;

	i = namedb_get(peerndx, targetfamily, scope);
	return i >= 0 ? &namedb[i].name : 0;
}

int NS_CLASS namedb_find(struct lunartarget_s *target)
{
	int i;

	if (!target || !target->family)
		return -1;

	for (i = 0; i < LUNAR_MAX_NAMES; i++) {
		if (target->family != namedb[i].name.family)
			continue;
		if (TARGET_EQUAL(&namedb[i].name, target))
			return i;
	}
	return -1;
}

int NS_CLASS namedb_add(struct lunartarget_s *target)
{
	int i;
	int f = -1;

	if (!target || !target->family)
		return -1;

	for (i = 0; i < LUNAR_MAX_NAMES; i++) {
		if (f == -1 && !namedb[i].name.family)
			f = i;
		if (target->family != namedb[i].name.family)
			continue;
		if (TARGET_EQUAL(&namedb[i].name,target)) {
			namedb_touch(i);
			return i;
		}
	}
	if (i >= LUNAR_MAX_NAMES) {
		if (f == -1)
			return -1;
		i = f;
	}
	memcpy(&namedb[i].name, target, sizeof(*target));
	namedb_touch(i);
	return i;
}

void NS_CLASS namedb_delpeer(int ndx)
{
	int i;

	for (i = 0; i < LUNAR_MAX_NAMES; i++) {
		if (namedb[i].name.family && namedb[i].peerndx == ndx) {
			namedb[i].name.family = 0;
			namedb[i].peerndx = -1;
		}
	}
}

void NS_CLASS namedb_touch(int ndx)
{
	namedb[ndx].expires = jiffies + LUNAR_DNS_TIMEOUT/1000000*HZ;
	DEBUG_TARGET(DBG_INFO, "namedb_touch:", &namedb[ndx].name);
}

// ----------------------------------------------------------------------


int NS_CLASS peer_getindex(struct lunartarget_s *target)
{
	int ndx;

	ndx = namedb_find(target);
	if (ndx >= 0)
		return namedb[ndx].peerndx;
	return -1;
}


int NS_CLASS peer_getnew(void)
{
	int i;

	DEBUG(DBG_PROC, printk("peer_getnew "));

	for (i = 0; i < LUNAR_MAX_PEERS; i++) {
		if (!peers[i].flags) {
#ifndef NS_PORT
			memset(peers + i, 0, sizeof(struct peer_s));
#endif
			peers[i].flags |= PEER_IS_VALID;
			DEBUG(DBG_PROC, printk("%d\n", i));
			return i;
		}
	}
	DEBUG(DBG_PROC, printk("-1\n"));
	return -1;
}


void NS_CLASS peer_set(int ndx, struct lunartarget_s *target)
{
	int namendx;

	DEBUG(DBG_PROC, printk("peer_set ndx=%d\n", ndx));

	namendx = namedb_add(target);
	if (namendx >= 0)
		namedb[namendx].peerndx = ndx;
}


void NS_CLASS peer_delpath(struct peer_s *n)
{
	int i;
	struct path_s *p = n->path;

	if (!p)
		return;

	DEBUG(DBG_PROC, printk("peer_delpath %d\n", n - peers));

	NETBOX_Instance netbox_deltimer(&(p->timer));
	p->used = 0;

	for (i = 0; i < LUNAR_MAX_PEERS; i++) {
		if (peers[i].path == p)
			peers[i].path = 0;
	}
	n->path = 0;
}


void NS_CLASS peer_del(struct peer_s *n)
{
	if (n->path) {
		SAPF_Instance sapf_deleteEntry(mysapfhandle,
					       n->path->fwdsel);
		peer_delpath(n);
	}
	namedb_delpeer(n-peers);
	n->flags = 0;
	//	NETBOX_Instance netbox_neigh_unmap(lunardev, &n->addr);
}


int NS_CLASS path_getnew(void)
{
	int i;

	for (i = 0; i < LUNAR_MAX_PATHS; i++) {
		if (!paths[i].used) {
#ifndef NS_PORT
			memset(paths + i, 0, sizeof(struct path_s));
#else
			paths[i].flags = 0;
#endif
			return i;
		}
	}
	return -1;
}


// ----------------------------------------------------------------------

/*   mkSapfDest
 *   allocates an eth+selector (netorder) buffer, to be used as a
 *   network pointer in the SAPF module. The SAPF module will free
 *   this netptr when the SAPF entry is removed.
 */
struct sapfdest_s* NS_CLASS mkSapfDest(unsigned char *eth, sapfsel_t sel)
{
	struct sapfdest_s *dst = (struct sapfdest_s*)
			kmalloc(sizeof(struct sapfdest_s), GFP_ATOMIC);
	char *cp;

	if (!dst)
		return 0;

	cp = eth ? (char*) eth : (char*) "\xff\xff\xff\xff\xff\xff";

	DEBUG(DBG_PROC, printk("LUNAR: mkSapfDest %s 0x%016Lx\n",
			 lunar_eth2str(cp), ntohsel(sel)));

	dst->ifnr = basedev ? basedev->ifindex : 0;
	dst->sel = sel;
	memcpy(dst->dst.sa_data, cp, ETH_ALEN);

	return dst;
}

/*   mkNetPointer
 *   installs a forwarder entry, returns the new selector.
 */
sapfsel_t NS_CLASS mkNetPointer(unsigned char *eth, sapfsel_t sel)
{
	sapfsel_t ptr = SAPF_Instance
		sapf_getUnusedSelector(mysapfhandle, SELCLASS_LOCAL);
        SAPF_Instance sapf_addCallback(mysapfhandle, ptr,
				       NS_CLASS_addr L2sendto,
				       mkSapfDest(eth, sel));
	return ptr;
}

sapfsel_t NS_CLASS mkHandlerPointer(sapfcallback_t fct, void *aux)
{
	sapfsel_t ptr = SAPF_Instance sapf_getUnusedSelector(mysapfhandle,
							SELCLASS_LOCAL);
        SAPF_Instance sapf_addCallback(mysapfhandle, ptr, fct, aux);
	return ptr;
}


// ----------------------------------------------------------------------

void NS_CLASS lunar_neigh_map(int ndx)
{
	// the low 48 bits of the fwd selector are mapped to an ethernet
	// address (all selectors are stored in nw order) which is then
	// associated with an IPv4 address (ARP table)

	struct lunartarget_s addr;
	int i = 0;

	memset(&addr, 0, sizeof(struct lunartarget_s));

	if (!peers[ndx].path)
		return;
	
	while( i > -1 )
	{
		i = namedb_getallipnames(ndx, i, &addr);
		
		if (i > -1) {
			NETBOX_Instance netbox_neigh_map(lunardev, &addr,
				((char*) &(peers[ndx].path->fwdsel)) +
					 sizeof(sapfsel_t) - ETH_ALEN);
		}
	}
}

int NS_CLASS peer_genaddr(struct lunartarget_s *peer4, struct lunartarget_s *peer6)
{
	struct lunartarget_s tmp, myself;
	struct in6_addr lunar_ipv6_prefix = LUNAR_IPV6_PREFIX;
	unsigned char rnd;
	
	memset( &tmp, 0, sizeof(struct lunartarget_s) );
	tmp.family = AF_INET;
	tmp.id.ipv4.s_addr = htonl(LUNAR_IP_SUBNET);	
	netbox_getL3addr(&myself, lunardev, AF_INET);
	
	do {
		get_random_bytes( &rnd, sizeof(char));
		tmp.id.ipv4.s_addr = htonl(LUNAR_IP_SUBNET + rnd);
	}
	while( namedb_find(&tmp) != -1 || TARGET_EQUAL(&myself, &tmp) || rnd == 0 || rnd == 255);
	
	memcpy( peer4, &tmp, sizeof(struct lunartarget_s) );
	DEBUG_TARGET(10, "LUNAR: generated addr4 ", &tmp);
	
	memset( &tmp, 0, sizeof(struct lunartarget_s) );
	tmp.family = AF_INET6;
	tmp.id.ipv6 = lunar_ipv6_prefix;
	myself = tmp;
	netbox_getL3addr(&myself, lunardev, AF_INET6);
		
	do {
		get_random_bytes( ((char *)&(tmp.id.ipv6)) + 8, 8);
	} 
	while( namedb_find(&tmp) != -1 || TARGET_EQUAL(&myself, &tmp) );
	
	memcpy( peer6, &tmp, sizeof(struct lunartarget_s) );
	DEBUG_TARGET(10, "LUNAR: generated addr6 ", &tmp);

	return 1;
}

int NS_CLASS peer_install(struct lunartarget_s *hit,
			  struct lunartarget_s *addr,
			  struct lunartarget_s *name)
{
	int ndx = -1;
	int h_ndx, a_ndx, n_ndx;

	if (!hit && !addr && !name)
		return -1;

	h_ndx = namedb_find(hit);
	a_ndx = namedb_find(addr); /* always -1 in this lunar version */
	n_ndx = namedb_find(name);

	DEBUG(DBG_INFO, printk("peer_install h=%d a=%d n=%d\n", h_ndx, a_ndx, n_ndx));

	if (h_ndx >= 0) {
	
		/* HIT is known */
	
		ndx = namedb[h_ndx].peerndx;
		
		if (n_ndx >= 0)
		{
			if (namedb[n_ndx].peerndx == ndx)
			{
				namedb_touch(h_ndx);  /* NAME is known and points to same peer */
				return ndx;
			}
			else
				return -1;            /* NAME is known but points to different peer */
		}
		else if (name)
		{	
			/* NAME is not known in namedb */
			/* --> we assume this is a new name */
			/* for the moment, if there exist an old name we replace it */

			int nndx = namedb_get( ndx, AF_NAME, 0);
			
			if (nndx > 0)
			{
				memcpy( &namedb[nndx].name, name, sizeof(struct lunartarget_s));
				return ndx;
			}
			else
			{
				n_ndx = namedb_add(name);
				if (n_ndx)
					namedb[n_ndx].peerndx = ndx;
					
				return ndx;
			}
		}
		else
			return -1;
		
	}

	if (n_ndx >= 0) {
	
		/* NAME is known but HIT is not known */
		
		/* TODO: check if this comes from a reboot */
		
		return -1;
	}
	
	if (!hit || !name)
		return -1;
	
	/* both HIT and NAME are unknown */
	
	ndx = peer_getnew();
	if (ndx < 0)
		return -1;
	
	h_ndx = namedb_add(hit);
	if (h_ndx >= 0)
		namedb[h_ndx].peerndx = ndx;
	
	n_ndx = namedb_add(name);
	if (n_ndx >= 0)
		namedb[n_ndx].peerndx = ndx;
	
	return ndx;
}


int NS_CLASS path_install(int ndx, sapfsel_t ptr, int plen, char timeroffs)
{
	unsigned long t;

	DEBUG(DBG_PROC, printk("LUNAR: install path %d 0x%016Lx\n",
			       ndx, ptr));

	if (!ptr)
		return -1;

	if (!peers[ndx].path) {
		int ndxp = path_getnew();
		if (ndxp < 0) {
//			peers[ndx].flags = 0;
			return -1;
		}
		peers[ndx].path = paths + ndxp;
	}

	peers[ndx].path->hops = plen;
	peers[ndx].path->flags &= ~PATH_IS_BIDIR;
	peers[ndx].path->fwdsel = ptr;
	peers[ndx].path->used = -1;

	t = (timeroffs ? conf_fri/2 : conf_fri) -
		((int) SAPF_Instance getRandomByte()-128) *
		conf_fri_jitter/128;

	NETBOX_Instance netbox_deltimer(&(peers[ndx].path->timer));
	NETBOX_Instance netbox_newtimer(&(peers[ndx].path->timer),
					_VOID_VOID_ path_timeout,
					peers + ndx, t);

//	if (!TARGET_IS_BROADCAST(&peers[ndx].addr))
		lunar_neigh_map(ndx);
	
	peers[ndx].path->flags |= PATH_IS_MAPPED;     // now resolved
	return ndx;
}

// ----------------------------------------------------------------------
//
void NS_CLASS unicast_rrep_handler(sapfpkt *p, void *ups_)
{
/*
	struct uc_pending_s *ups = (struct uc_pending_s *)ups_;
	struct xrp_head_s *x = (struct xrp_head_s*) p->data;
	struct xrp_sel_eth_s *f;
	int ndx;

        DEBUG(DBG_LUNAR, printk("LUNAR: unicast_rrep_handler (%d bytes)\n",
				p->len));

	if (p->len < sizeof(struct xrp_head_s) ||
	    ntohs(x->cmd) != XRP_CMD_LUNAR_RREP) {
		DEBUG(DBG_INFO, printk("  unknown or garbled RREP\n"));
		return ;
	}
*/
	struct uc_pending_s *ups = (struct uc_pending_s *)ups_;
	struct xrp_head_s *x = (struct xrp_head_s *) p->data;
	xrp_param_t *pp; // pointer to a single parameter block
	struct xrp_sel_eth_s *fwd = 0;
	int cnt, len;
	int pndx, nndx;
	struct lunartarget_s s_hit, s_addr, s_addr6, s_name, *addr = 0;
	struct lunartarget_s *srchit = 0, *srcaddr = 0, *srcaddr6 = 0, *srcname = 0;
	struct lunartarget_s peer4, peer6;
	
        DEBUG(DBG_LUNAR, printk("LUNAR: unicast_rrep_handler (%d bytes)\n",
				p->len));

	if (p->len < sizeof(struct xrp_head_s) ||
	    ntohs(x->cmd) != XRP_CMD_LUNAR_RREP) {
		DEBUG(DBG_INFO, printk(" wrong XRP cmd (not a RREP)\n"));
		return;
	}
	// parse RREP:
	len = p->len - sizeof(*x);
	pp = (xrp_param_t *) (x+1);
	while (len >= 2) {
		if (ntohs(pp->param.len) & XRP_CMD_BIT)
			break;
		if (len < 4 ||
		    ntohs(pp->param.len) > len ||
		    ntohs(pp->param.len) < 4) {
			DEBUG(DBG_INFO, printk(" end of XRP parsing: wrong length %d\n", len));
			return;
		}
		if (pp->param.partype > XRP_CTYPE_UNKNOWN &&
		    pp->param.partype < XRP_CTYPE__COUNT) {
			if (pp->param.partype != XRP_CTYPE_NAME &&
			    ntohs(pp->param.len) !=
			    		xrp_param_len[pp->param.partype]) {
				DEBUG(DBG_INFO, printk(" wrong parameter length. class=%d, type=%d, len=%d, explen=%d\n", pp->param.parclass, pp->param.partype, ntohs(pp->param.len), pp->param.partype));
				return;
			}
			DEBUG(DBG_ALL, printk(" param class+len+expect_len: %d %d %d\n",
			pp->param.parclass, ntohs(pp->param.len), xrp_param_len[pp->param.partype]));

			switch (pp->param.parclass) {
			case XRP_CLASS_LUNAR_RREP_FWDPTR:
				if (pp->param.partype == XRP_CTYPE_SEL_ETH)
					fwd = (struct xrp_sel_eth_s *) pp;
				break;
			case XRP_CLASS_LUNAR_RREP_SRC:
				switch(pp->param.partype) {
				case XRP_CTYPE_IPv4:
					if (!param2target(&s_addr, pp)) {
						srcaddr = &s_addr;
						DEBUG_TARGET(10, "  got IP",
							     srcaddr);
					}
					break;
#ifdef LUNAR_HAS_IPV6
				case XRP_CTYPE_IPv6:
					if (!param2target(&s_addr6, pp)) {
						srcaddr6 = &s_addr6;
						DEBUG_TARGET(10, "  got IP6",
							     srcaddr6);
					}
					break;
#endif
				case XRP_CTYPE_NAME:
					if (!param2target(&s_name, pp)) {
						srcname = &s_name;
						DEBUG_TARGET(10, "  got name",
							     srcname);
					}
					break;
				case XRP_CTYPE_HIT128:
					if (!param2target(&s_hit, pp)) {
						srchit = &s_hit;
						DEBUG_TARGET(10, "  got hit",
							     srchit);
					}
					break;
				default:
					break;
				}
				break;
			default: // ignore unknown fields
				break;
			}
		}
		cnt = (ntohs(pp->param.len) + 3) & ~0x3;
		len -= cnt;
		pp = (xrp_param_t*) (((char*)pp) + cnt);
	}

	if (!fwd) {
		DEBUG(DBG_INFO, printk("  no forward addr found in RREP\n"));
		return ;
	}

	pndx = peer_getindex(&ups->target);
	if (pndx < 0) {
		DEBUG(DBG_INFO, printk("  target not in peers anymore\n"));
		return;
	}
	
	// IP-less LUNAR: if IP addresses are unknown then generate them
	
	if ( !(peers[pndx].flags & PEER_IS_FAKE) &&
	    (!namedb_getname(pndx, AF_INET6, IPV6_ADDR_SCOPE_MANET) ||
	     !namedb_getname(pndx, AF_INET, 0)) )
	{
		peer_genaddr( &peer4, &peer6 );
		peers[pndx].flags |= PEER_IS_FAKE;
		
		nndx = namedb_add(&peer4);
		if (nndx >= 0)
		{
			namedb[nndx].peerndx = pndx;
			srcaddr = &peer4;
		}
			
		nndx = namedb_add(&peer6);
		if (nndx >= 0)
		{
			namedb[nndx].peerndx = pndx;
			srcaddr6 = &peer6;
		}
	}
	
	/* 2005-06-30 CJ: added merging support
	 *
	 * i.e. if peer already existed but could not be identified 
	 * (e.g. ping IPv6 addr when only name, HIT, and IPv4 were known)
	 * --> we merge the two peers
	*/
	
	nndx = namedb_find(srchit);
	if (nndx >= 0) {
		if (namedb[nndx].peerndx != pndx ||
			!TARGET_EQUAL(srchit, &namedb[nndx].name)) {
			DEBUG(DBG_INFO, printk("  --> doing peer merging\n"));
			
			namedb[ups->ndx].peerndx = namedb[nndx].peerndx; /* point to old peer */
			namedb_touch(ups->ndx);
			namedb_touch(nndx);
			
			peer_del(&peers[pndx]);		/* not needed any more  */
			pndx = namedb[nndx].peerndx;	/* for rest of the code */
		}
	
	} else if (srchit) {
		
		nndx = namedb_add(srchit);
		if (nndx >= 0)
			namedb[nndx].peerndx = pndx;
	}

	nndx = namedb_find(srcaddr);
	if (nndx >= 0) {
		if (namedb[nndx].peerndx != pndx ||
			!TARGET_EQUAL(srcaddr, &namedb[nndx].name))
			DEBUG(DBG_INFO, printk("  suspicious RREP (ADDR mismatch)\n"));
	} else if (srcaddr) {
		nndx = namedb_add(srcaddr);
		if (nndx >= 0)
			namedb[nndx].peerndx = pndx;
	} 

#ifdef LUNAR_HAS_IPV6
	if (!IamIPv6Ready()) {
		srcaddr6 = 0;
		goto skipv6;	// we are not IPv6 enabled !
	}
	
	nndx = namedb_find(srcaddr6);
	if (nndx >= 0) {
		if (namedb[nndx].peerndx != pndx ||
			!TARGET_EQUAL(srcaddr6, &namedb[nndx].name))
			DEBUG(DBG_INFO, printk("  suspicious RREP (ADDR6 mismatch)\n"));
	} else if (srcaddr6) {
		nndx = namedb_add(srcaddr6);
		if (nndx >= 0)
			namedb[nndx].peerndx = pndx;
	}
	
skipv6:
#endif

#ifdef LUNAR_HAS_DNS
	nndx = namedb_find(srcname);
	if (nndx >= 0) {
		if (namedb[nndx].peerndx != pndx ||
		    !TARGET_EQUAL(srcname, &namedb[nndx].name)) {
			DEBUG(DBG_INFO, printk("  suspicious RREP (NAME mismatch)\n"));
			nndx = -1;
		}
	} else if (srcname) {
		nndx = namedb_add(srcname);
		if (nndx >= 0)
			namedb[nndx].peerndx = pndx;
	} 
#endif

	path_install(pndx, mkNetPointer(fwd->eth, fwd->sel), x->ttl, 0);

	if (ups->is_bidir)
	{
		// SAPF_Instance sapf_addREDIR(mysapfhandle, ups->backsel, ipv4deliver);
		
		SAPF_Instance sapf_addCallback(mysapfhandle, ups->backsel, netbox_nat, peers + pndx);
	}

#ifdef LUNAR_HAS_DNS
	if (ups->udp_port && ups->trans_id && ups->ndx >= 0) {
		DEBUG(DBG_INFO, printk("  sending DNS reply to IP stack\n"));
		
		if (ups->qnamefamily == AF_INET)
			addr = srcaddr;
#ifdef LUNAR_HAS_IPV6
		else
			addr = srcaddr6;
#endif
		if ( ups->ndx > -1 && addr)	// original query index in namedb
			send_dns_reply(&namedb[ups->ndx].name, addr,
			       ups->udp_port, ups->trans_id, NOT_PTR);
		
		else if ( ups->ndx > -1 && !addr)	// name unresolved
			send_dns_nosuchname(&namedb[ups->ndx].name,
				ups->udp_port, ups->trans_id, ups->qnamefamily);
	}
#endif

	NETBOX_Instance netbox_deltimer(&(peers[pndx].timer));
	NETBOX_Instance netbox_newtimer(&(peers[pndx].timer),
						_VOID_VOID_ peer_timeout,
						&peers[pndx], LUNAR_DNS_TIMEOUT);
	
	memset( &(peers[pndx].upssel), 0, sizeof(sapfsel_t));
        SAPF_Instance sapf_deleteEntry(mysapfhandle, ups->caller);

        return ;
}

void NS_CLASS unicast_transit_rrep_handler(sapfpkt *p,void *tr_)
{
  struct uc_transit_pending_s *tr =(struct uc_transit_pending_s *)tr_;
	int len, ifnr = basedev ? basedev->ifindex : 0;
	struct xrp_head_s *x;
	struct xrp_sel_eth_s *f;
	struct sockaddr sa;
	char buf[300];

        DEBUG(DBG_LUNAR, printk("LUNAR: unicast transit rrep handler (%d bytes)\n",
			p->len));

	if (p->len < sizeof(struct xrp_head_s))
		return ;

	len = MIN(sizeof(buf), p->len);
	memcpy(buf, p->data, len);
	x = (struct xrp_head_s *) buf;
	if (ntohs(x->cmd) != XRP_CMD_LUNAR_RREP) {
		DEBUG(DBG_INFO, printk("  unknown or garbled RREP\n"));
		return ;
	}
	x->ttl++;

	f = (struct xrp_sel_eth_s *)
		xrp_locate(x+1, len - sizeof(*x), XRP_CLASS_LUNAR_RREP_FWDPTR);
	if (!f || f->partype != XRP_CTYPE_SEL_ETH) {
		DEBUG(DBG_INFO, printk("  no forward addr found in RREP %p\n", f));
		return ;
	}

	if (tr->is_bidir)
		SAPF_Instance sapf_addCallback(mysapfhandle, tr->mybacksel, NS_CLASS_addr L2sendto,
				 mkSapfDest(tr->backeth, tr->backsel));

	NETBOX_Instance netbox_getL2addr(ifnr, &sa);
	f->sel = mkNetPointer(f->eth, f->sel);
	memcpy(f->eth, sa.sa_data, ETH_ALEN);

	memcpy(sa.sa_data, tr->replyeth, ETH_ALEN);
	NETBOX_Instance netbox_l2send(ifnr, &sa, tr->replysel, buf, len);

        SAPF_Instance sapf_deleteEntry(mysapfhandle, tr->caller);

        return ;
}


void NS_CLASS bcast_rep_handler(sapfpkt *p, void *bps_)
{
	struct bcast_pending_s *bps=(struct bcast_pending_s *)bps_;
	struct xrp_head_s *x = (struct xrp_head_s*) p->data;
	struct xrp_sel_eth_s *f;

        DEBUG(DBG_PROC, printk("bcast_rep_handler (%d bytes)\n", p->len));

	if (p->len < sizeof(struct xrp_head_s) ||
	    ntohs(x->cmd) != XRP_CMD_LUNAR_RREP) {
		DEBUG(DBG_INFO, printk("  unknown or garbled REP\n"));
		return ;
	}
	f = (struct xrp_sel_eth_s *)xrp_locate(x+1, p->len - sizeof(*x), XRP_CLASS_LUNAR_RREP_FWDPTR);
	if (!f || f->partype != XRP_CTYPE_SEL_ETH) {
		DEBUG(DBG_INFO, printk("  no forward addr found in RREP %p\n", f));
		return ;
	}
/*
	if (f->sel != bps->fwd_sel) {
		DEBUG(DBG_INFO, printk("  returned bc sel does not match\n"));
		return ;
	}
*/

	bc_sel = mkNetPointer(0, bps->fwd_sel);
	if (!bc_len) {
		path_install(bps->ndx, bc_sel, 3, 0);
//		install_peer(0, &bps->target, bc_sel, 3, 0);
		if (bc_selp)
			*bc_selp = bc_sel;
	}

	SAPF_Instance sapf_deleteEntry(mysapfhandle, bps->caller);
       
	return ;
}

// ----------------------------------------------------------------------

int NS_CLASS param2target(struct lunartarget_s *t, xrp_param_t *p)
{
	switch (p->param.partype) {
	case XRP_CTYPE_IPv4:
		t->family = AF_INET;
		t->id.ipv4 = p->ipv4.ip;
		break;
#ifdef LUNAR_HAS_IPV6
	case XRP_CTYPE_IPv6:
		t->family = AF_INET6;
		memcpy(&(t->id.ipv6), &(p->ipv6.ip6), 16);
		break;
#endif
#ifdef LUNAR_HAS_DNS
	case XRP_CTYPE_NAME:
		t->family = AF_NAME;
		strncpy(t->id.name, p->name.name, XRP_MAX_NAME_LEN);
		t->id.name[XRP_MAX_NAME_LEN] = '\0';
		break;
#endif
	case XRP_CTYPE_HIT128:
		t->family = AF_HIT;
		memcpy(t->id.hit, p->hit128.hit, sizeof(hit_t));
		break;
	default:
		DEBUG(DBG_INFO, printk(" unknown target type %d\n",
				p->param.partype));
		return 1;
	}
	return 0;
}


void NS_CLASS lunar_rreq_handler_bc(struct lunartarget_s *target,
				    struct xrp_sel_s *series,
				    struct xrp_sel_eth_s *reply,
				    struct xrp_sel_s *fwd,
				    sapfpkt *p)
{
/*
		unsigned char buf[300], *cp = buf;
		sapfsel_t ptr;
*/
	int ifnr = basedev ? basedev->ifindex : 0;
	struct bc_fanout_s *bf;

	DEBUG(DBG_LUNAR, printk("broadcast tree request received\n"));

	if (!fwd) {
		DEBUG(DBG_INFO, printk("  no fwd sel\n"));
		return;
	}
	bf = kmalloc(sizeof(struct bc_fanout_s), GFP_ATOMIC);
	if (!bf)
		return;

	bf->local = ipv4deliver;
	memset(bf->eth, 0xff, ETH_ALEN);
	bf->forward = SAPF_Instance sapf_getUnusedSelector(mysapfhandle,
						       SELCLASS_REMOTE);
	sapf_addCallback(mysapfhandle, fwd->sel, lunar_bcast_fanout, bf);
	fwd->sel = bf->forward;
	NETBOX_Instance netbox_l2send(ifnr, 0,
				      htonsel(SAPF_SELNET_LUNAR_RREQ),
				      p->data, p->len);

/*
		ptr = SAPF_Instance sapf_getUnusedSelector(mysapfhandle, fwd->sel);
		SAPF_Instance sapf_addREDIR(mysapfhandle, ptr, ipv4deliver);

		cp += xrp_fill_hdr(cp, XRP_CMD_LUNAR_RREP, 0);
		cp += xrp_fill_sel_eth(cp, XRP_CLASS_LUNAR_RREP_FWDPTR, ptr, (unsigned char *)sa.sa_data);

		if (!NETBOX_Instance netbox_getHIT128(&hit))
			cp += xrp_fill_hit128(cp, XRP_CLASS_LUNAR_RREQ_SOURCEHIT, &hit);
		cp += xrp_fill_eom(cp);

		memcpy(sa.sa_data, reply->eth, ETH_ALEN);
		NETBOX_Instance netbox_l2send(ifnr, &sa, reply->sel, buf, cp - buf);

		return;
*/
}


void NS_CLASS lunar_rreq_handler_uc(int ifnr, struct sockaddr *eth,
				    struct xrp_sel_eth_s *reply,
				    struct xrp_sel_eth_s *backwd,
				    struct xrp_sel_s *fwd,
				    struct lunartarget_s *target,
				    struct lunartarget_s *srchit,
				    struct lunartarget_s *srcaddr,
				    struct lunartarget_s *srcname)
{
	sapfsel_t ptr;
	unsigned char buf[300], *cp = buf;
	int ndx = -1, nndx = -1;

	struct lunartarget_s peer4, peer6;
	
	ndx = peer_install(srchit, 0, srcname);
	
	if (ndx < 0)
	{
		DEBUG(DBG_INFO, printk("rreq_handler_uc: peer_install failed\n"));
		return;
	}
	else {
		NETBOX_Instance netbox_deltimer(&(peers[ndx].timer));
		NETBOX_Instance netbox_newtimer(&(peers[ndx].timer),
						_VOID_VOID_ peer_timeout,
						&peers[ndx], LUNAR_DNS_TIMEOUT);
	}
	
	
	// IP-less LUNAR: create faked IP address for correspondent
	
	if ( !(peers[ndx].flags & PEER_IS_FAKE) &&
	     ( !namedb_getname(ndx, AF_INET6, IPV6_ADDR_SCOPE_MANET) ||
	       !namedb_getname(ndx, AF_INET, 0) ))
	{
		peers[ndx].flags |= PEER_IS_FAKE;
	
		peer_genaddr( &peer4, &peer6 );
		
		nndx = namedb_add(&peer4);
		if (nndx >= 0)
			namedb[nndx].peerndx = ndx;
					
		nndx = namedb_add(&peer6);
		if (nndx >= 0)
			namedb[nndx].peerndx = ndx;
	}
	
	
	if (ndx >= 0 && backwd) {
		if (peers[ndx].path &&
		    (peers[ndx].path->flags & PATH_IS_MAPPED)) {
			struct lunartarget_s *t;
			// known src with a path: delete it before update
			peers[ndx].path->flags &= ~PATH_IS_MAPPED;
			SAPF_Instance sapf_deleteEntry(mysapfhandle,
					       peers[ndx].path->fwdsel);
			t = namedb_getname(ndx, AF_INET, 0);
			if (t)
			    NETBOX_Instance netbox_neigh_unmap(lunardev, t);
#ifdef LUNAR_HAS_IPV6
			t = namedb_getname(ndx, AF_INET6, IPV6_ADDR_SCOPE_MANET);
			if (t)
			    NETBOX_Instance netbox_neigh_unmap(lunardev, t);
#endif
		}
		path_install(ndx, mkNetPointer(backwd->eth, backwd->sel),
			     9, 1);
	} else
		DEBUG(DBG_INFO, printk("rreq_handler_uc: unidirectional path setup\n"));

	ptr = SAPF_Instance sapf_getUnusedSelector(mysapfhandle, SELCLASS_LOCAL);
	// SAPF_Instance sapf_addREDIR(mysapfhandle, ptr, ipv4deliver);
	
	SAPF_Instance sapf_addCallback(mysapfhandle, ptr, netbox_nat, peers + ndx);

	
	cp += xrp_fill_hdr(cp, XRP_CMD_LUNAR_RREP, 0);
	cp += xrp_fill_sel_eth(cp, XRP_CLASS_LUNAR_RREP_FWDPTR,
				       ptr, (unsigned char*)eth->sa_data);

	if (srchit || srcaddr || srcname) {
		// if any source details were given, we tell everything
		// except for the target we were found for
		struct lunartarget_s t;

		/* IP-less LUNAR forbids IP addresses to be shared 

		if (!TARGET_IS_IPV4(target)) {
			NETBOX_Instance netbox_getL3addr(&t, lunardev,
							 AF_INET);
			cp += xrp_fill_ipv4(cp, XRP_CLASS_LUNAR_RREP_SRC,
					    t.id.ipv4);
		}
#ifdef LUNAR_HAS_IPV6
		if (!TARGET_IS_IPV6(target)) {
			// MANET scope, or local scope for gateways
			// (gateways do not always autoconfigure a MANET-local address)
			memset(&t, 0, sizeof(struct lunartarget_s));
			memset(&t, 0xfd, 1); 
			NETBOX_Instance netbox_getL3addr(&t, lunardev,
							 AF_INET6);
							 
			if (!IN6_IS_ADDR_UNSPECIFIED(&t.id.ipv6))
				cp += xrp_fill_ipv6(cp, XRP_CLASS_LUNAR_RREP_SRC,
					    &t.id.ipv6);
		}
#endif
		*/	

#ifdef LUNAR_HAS_DNS
		if (!TARGET_IS_NAME(target)) {
			NETBOX_Instance netbox_getL3addr(&t, lunardev,
							 AF_NAME);
			cp += xrp_fill_name(cp, XRP_CLASS_LUNAR_RREP_SRC,
					    t.id.name);
		}
#endif
		if (!TARGET_IS_HIT(target)) {
			NETBOX_Instance netbox_getL3addr(&t, lunardev,
							 AF_HIT);
			cp += xrp_fill_hit128(cp, XRP_CLASS_LUNAR_RREP_SRC,
					      &t.id.hit);
		}
	}
	cp += xrp_fill_eom(cp);

	memcpy(eth->sa_data, reply->eth, ETH_ALEN);
	NETBOX_Instance netbox_l2send(ifnr, eth, reply->sel, buf, cp - buf);
}


void NS_CLASS lunar_rreq_handler_fwd(sapfpkt *p,
				     struct xrp_sel_eth_s *reply,
				     struct xrp_sel_eth_s *backwd,
				     struct sockaddr *eth)
{
	// re-broadcast the request
	struct uc_transit_pending_s *tr;

	tr = (struct uc_transit_pending_s *)kmalloc(sizeof(*tr), GFP_ATOMIC);
	if (!tr)
		return;
	memset(tr, 0, sizeof(*tr));

	tr->replysel = reply->sel;
	memcpy(tr->replyeth, reply->eth, ETH_ALEN);

	// we reuse the RREQ packet and replace fields:
	memcpy(reply->eth, eth->sa_data, ETH_ALEN);
	tr->caller = reply->sel = mkHandlerPointer(
			NS_CLASS_addr unicast_transit_rrep_handler, tr);
	if (backwd) {
		tr->is_bidir = 1;
		tr->backsel = backwd->sel;
		memcpy(tr->backeth, backwd->eth, ETH_ALEN);

		tr->mybacksel = backwd->sel =
			SAPF_Instance sapf_getUnusedSelector(mysapfhandle,
							   SELCLASS_LOCAL);
		memcpy(backwd->eth, eth->sa_data, ETH_ALEN);
	}

	schedule_rreq(p->data, p->len, 0);
}


void NS_CLASS lunar_rreq_handler(sapfpkt *p, void *aux)
{
	struct xrp_head_s *x = (struct xrp_head_s *) p->data;
	struct xrp_sel_eth_s *reply = 0, *backwd = 0;
	struct xrp_sel_s *series = 0, *fwd = 0;
	xrp_param_t *pp; // pointer to a single parameter block
	xrp_param_t *target = 0;
	int cnt, len;
	struct lunartarget_s lunar_target;
	struct lunartarget_s s_hit, s_addr, s_name;
	struct lunartarget_s *srchit = 0, *srcaddr = 0, *srcname = 0;

	DEBUG(DBG_LUNAR, printk("LUNAR: received RREQ (%d bytes)\n", p->len));

	if (p->len < sizeof(struct xrp_head_s) ||
	    ntohs(x->cmd) != XRP_CMD_LUNAR_RREQ) {
		DEBUG(DBG_INFO, printk(" wrong XRP cmd (not a RREQ)\n"));
		return;
	}
	// parse RREQ:
	len = p->len - sizeof(*x);
	pp = (xrp_param_t *) (x+1);
	while (len >= 2) {
		if (ntohs(pp->param.len) & XRP_CMD_BIT)
			break;
		if (len < 4 ||
		    ntohs(pp->param.len) > len ||
		    ntohs(pp->param.len) < 4) {
			DEBUG(DBG_INFO, printk(" end of XRP parsing: wrong length %d\n", len));
			return;
		}
		if (pp->param.partype > XRP_CTYPE_UNKNOWN &&
		    pp->param.partype < XRP_CTYPE__COUNT) {
			if (pp->param.partype != XRP_CTYPE_NAME &&
			    ntohs(pp->param.len) !=
			    		xrp_param_len[pp->param.partype]) {
				DEBUG(DBG_INFO, printk(" wrong parameter length. class=%d, type=%d, len=%d, explen=%d\n", pp->param.parclass, pp->param.partype, ntohs(pp->param.len), pp->param.partype));
				return;
			}
			DEBUG(DBG_ALL, printk(" param class+len+expect_len: %d %d %d\n",
			pp->param.parclass, ntohs(pp->param.len), xrp_param_len[pp->param.partype]));

			switch (pp->param.parclass) {
			case XRP_CLASS_LUNAR_RREQ_SERIES:
				if (pp->param.partype == XRP_CTYPE_SELECTOR)
					series = (struct xrp_sel_s *) pp;
				break;
			case XRP_CLASS_LUNAR_RREQ_REPLYTO:
				if (pp->param.partype == XRP_CTYPE_SEL_ETH)
					reply = (struct xrp_sel_eth_s *) pp;
				break;
			case XRP_CLASS_LUNAR_RREQ_TARGET:
				target = (xrp_param_t *) pp;
				break;
			case XRP_CLASS_LUNAR_RREQ_BACKPTR:
				if (pp->param.partype == XRP_CTYPE_SEL_ETH)
					backwd = (struct xrp_sel_eth_s *) pp;
				break;
			case XRP_CLASS_LUNAR_RREQ_FWDPTR: // for broadcast
				if (pp->param.partype == XRP_CTYPE_SELECTOR)
					fwd = (struct xrp_sel_s *) pp;
				break;
			case XRP_CLASS_LUNAR_RREQ_SRC:
				switch(pp->param.partype) {
				case XRP_CTYPE_IPv4:
				case XRP_CTYPE_IPv6:
					if (!param2target(&s_addr, pp))
						srcaddr = &s_addr;
					break;
				case XRP_CTYPE_NAME:
					if (!param2target(&s_name, pp))
						srcname = &s_name;
					break;
				case XRP_CTYPE_HIT128:
					if (!param2target(&s_hit, pp))
						srchit = &s_hit;
					break;
				default:
					break;
				}
				break;
			default: // ignore unknown fields
				break;
			}
		}
		cnt = (ntohs(pp->param.len) + 3) & ~0x3;
		len -= cnt;
		pp = (xrp_param_t*) (((char*)pp) + cnt);
	}

	// prevent broadcast storms:
	if (!series || SAPF_Instance sapf_exists(mysapfhandle, series->sel)) {
		DEBUG(DBG_INFO, printk(series ? "  series %016Lx already exists, dropping\n" : "  no series\n", series ? ntohsel(series->sel) : 0L));
		return;
	}
	SAPF_Instance sapf_addREDIR(mysapfhandle, series->sel, 0);

	if (!target || !reply || param2target(&lunar_target, target)) {
		DEBUG(DBG_INFO, printk("  no target, no reply field or invalid target\n"));
		return;
	}

	DEBUG_TARGET(DBG_INFO, "RREQ target is", &lunar_target);

	if (TARGET_IS_BROADCAST(&lunar_target))
		lunar_rreq_handler_bc(&lunar_target, series, reply, fwd, p);
	else {
		struct lunartarget_s myself;
		struct sockaddr eth;
		int ifnr = basedev ? basedev->ifindex : 0;

		NETBOX_Instance netbox_getL2addr(ifnr, &eth);
		myself = lunar_target;
		NETBOX_Instance netbox_getL3addr(&myself, lunardev,
						 lunar_target.family);
		DEBUG_TARGET(DBG_INFO, "  my addr is", &myself);

/*
#ifdef LUNAR_HAS_IPV6
		if (TARGET_EQUAL(&lunar_target, &myself) && srcaddr && TARGET_IS_IPV6(srcaddr)) {
			// am I IPv6 enabled ?
			if( !IamIPv6Ready() ) {
				DEBUG(DBG_LUNAR, printk("LUNAR: Discovered but no such family\n"));
				if (srcname) {
					int ndx;					
					
					if ((ndx = namedb_find(srcname)) >= 0)
						namedb_touch(ndx);
					else {
					
						ndx = peer_install(srchit, 0, srcname);
						if (ndx > -1)
							NETBOX_Instance netbox_newtimer(&(peers[ndx].timer),
							_VOID_VOID_ peer_timeout,
							&peers[ndx], LUNAR_DNS_TIMEOUT);
					}
				}
				return;
			}
		}
#endif */
		if (TARGET_EQUAL(&lunar_target, &myself)) {
			DEBUG(DBG_LUNAR, printk("LUNAR: I was discovered\n"));
			lunar_rreq_handler_uc(ifnr, &eth,
					      reply, backwd, fwd,
					      &lunar_target,
					      srchit, 0, srcname);
		} else {
		
			if (srcname) {
				int ndx;
				
				if ((ndx = namedb_find(srcname)) >= 0)
						namedb_touch(ndx);
				else {
					ndx = peer_install(srchit, 0, srcname);
					if (ndx > -1)
						NETBOX_Instance netbox_newtimer(&(peers[ndx].timer),
						_VOID_VOID_ peer_timeout,
						&peers[ndx], LUNAR_DNS_TIMEOUT);
				}
			}
												
			if (x->ttl-- > 1)
				lunar_rreq_handler_fwd(p, reply, backwd, &eth);
		}
	}
}


void NS_CLASS send_rreq(int radius, sapfsel_t series,
			struct lunartarget_s *target,
			sapfsel_t reply, sapfsel_t backsel,
			struct lunartarget_s *srchit,
			struct lunartarget_s *srcaddr,
			struct lunartarget_s *srcname,
			sapfsel_t fwd)
{
	struct sockaddr sa;
	int ifnr = basedev ? basedev->ifindex : 0;
	char buf[300], *cp = buf;

	cp = buf;
	cp += xrp_fill_hdr(cp, XRP_CMD_LUNAR_RREQ, radius);
	cp += xrp_fill_sel(cp, XRP_CLASS_LUNAR_RREQ_SERIES, series);
	NETBOX_Instance netbox_getL2addr(ifnr, &sa);
	cp += xrp_fill_sel_eth(cp, XRP_CLASS_LUNAR_RREQ_REPLYTO,
			       reply, (unsigned char *)sa.sa_data);

	switch (target->family) {
	case AF_INET:
		cp += xrp_fill_ipv4(cp, XRP_CLASS_LUNAR_RREQ_TARGET,
				    target->id.ipv4);
		break;
#ifdef LUNAR_HAS_IPV6
	case AF_INET6:
		cp += xrp_fill_ipv6(cp, XRP_CLASS_LUNAR_RREQ_TARGET,
				    &target->id.ipv6);
		break;
#endif
#ifdef LUNAR_HAS_DNS
	case AF_NAME:
		cp += xrp_fill_name(cp, XRP_CLASS_LUNAR_RREQ_TARGET,
				    target->id.name);
		break;
#endif
	case AF_HIT:
		cp += xrp_fill_hit128(cp, XRP_CLASS_LUNAR_RREQ_TARGET,
				    &target->id.hit);
		break;
	default:
		break;
	}
	DEBUG_TARGET(DBG_INFO, "send_rreq TARGET ", target);

	if (srchit) {
		cp += xrp_fill_hit128(cp, XRP_CLASS_LUNAR_RREQ_SRC,
				      &srchit->id.hit);
		DEBUG_TARGET(DBG_INFO, "send_rreq SRCHIT ", srchit);
	}
	
	/* For IP-less LUNAR we do not send source address any more 
		
	if (srcaddr && !TARGET_EQUAL(srcaddr, target)) {
		if( TARGET_IS_IPV4( srcaddr ))
			cp += xrp_fill_ipv4(cp, XRP_CLASS_LUNAR_RREQ_SRC,
					    srcaddr->id.ipv4);
		
#ifdef LUNAR_HAS_IPV6
		else if( TARGET_IS_IPV6( srcaddr ))
			cp += xrp_fill_ipv6(cp, XRP_CLASS_LUNAR_RREQ_SRC,
					    &srcaddr->id.ipv6);
#endif		
		DEBUG_TARGET(DBG_INFO, "send_rreq SRCADDR", srcaddr);
	} */

#ifdef LUNAR_HAS_DNS
	if (srcname) {
		cp += xrp_fill_name(cp, XRP_CLASS_LUNAR_RREQ_SRC,
				    srcname->id.name);
		DEBUG_TARGET(DBG_INFO, "send_rreq SRCNAME", srcname);
		
		/* because we are sending RREQ with our name */
		/* we postpone the tx of name HELLO messages */
		
		NETBOX_Instance netbox_deltimer(&nametimer);
		NETBOX_Instance netbox_newtimer(&nametimer, 
				_VOID_VOID_ lunar_send_name, 0, LUNAR_NAME_TIMEOUT);
	}
#endif

	if (backsel) {
		cp += xrp_fill_sel_eth(cp, XRP_CLASS_LUNAR_RREQ_BACKPTR,
			      	 backsel, (unsigned char *)sa.sa_data);
	}

	if (fwd) {
		cp += xrp_fill_sel(cp, XRP_CLASS_LUNAR_RREQ_FWDPTR, fwd);
	}

	cp += xrp_fill_eom(cp);

	NETBOX_Instance netbox_l2send(ifnr, 0,
				      htonsel(SAPF_SELNET_LUNAR_RREQ),
				      buf, cp - buf);
}


void NS_CLASS start_rreq(int peerndx, int ndx, int udp_port, int trans_id, char namefamily)
{
	sapfsel_t series;
	int radius;
	long tout;
	struct lunartarget_s *t;
	struct bcast_pending_s *bps = 0;
	struct uc_pending_s *ups = 0;

	t = namedb_getname(peerndx, AF_HIT, 0);
	if (!t && ndx>-1)
		t = &namedb[ndx].name;	// if we cannot find HIT, we use namedb entry
	
/*	if (!t)
		t = namedb_getname(peerndx, AF_INET, 0);
#ifdef LUNAR_HAS_IPV6
	if (!t)
		t = namedb_getname(peerndx, AF_INET6, trans_id);// trans_id is IPv6 scope
#endif
#ifdef LUNAR_HAS_DNS
	if (!t)
		t = namedb_getname(peerndx, AF_NAME, 0);
#endif */
	
	if (!t) {
		DEBUG(DBG_LUNAR, printk("LUNAR: start_rreq, no valid target!?\n"));
		return;
	}

	DEBUG_TARGET(DBG_LUNAR, "LUNAR: start_rreq called for", t);
	DEBUG(DBG_LUNAR, printk("       udp_port=%d, trans_id=%d\n",
		     ntohs(udp_port), ntohs(trans_id)));

	if (!peers[peerndx].path)
		return;

	radius = MIN(peers[peerndx].path->used, conf_max_ring);
	series = SAPF_Instance sapf_getUnusedSelector(mysapfhandle, SELCLASS_REMOTE);
	SAPF_Instance sapf_addREDIR(mysapfhandle, series, 0);

	if (!TARGET_IS_BROADCAST(t)) {
		struct lunartarget_s myhit;
		NETBOX_Instance netbox_getL3addr(&myhit, lunardev, AF_HIT);
		
		ups = (struct uc_pending_s *)kmalloc(sizeof(struct uc_pending_s), GFP_ATOMIC);
		if (!ups) {
			peer_delpath(peers + peerndx);
			return;
		}
		memset(ups, 0, sizeof(struct uc_pending_s));
		ups->target = *t;
		ups->caller = mkHandlerPointer(
			NS_CLASS_addr unicast_rrep_handler, ups);
		ups->udp_port = udp_port;
		ups->trans_id = trans_id;
		ups->ndx = ndx;	// namedb entry that triggered RREQ
		ups->qnamefamily = namefamily;
		peers[peerndx].upssel = ups->caller;
				
		if (peers[peerndx].path->flags & PATH_IS_BIDIR) {
			struct lunartarget_s myaddr;
#ifdef LUNAR_HAS_DNS
			struct lunartarget_s myname;
			NETBOX_Instance netbox_getL3addr(&myname, lunardev, AF_NAME);
#endif
			memset(&myaddr, 0, sizeof(myaddr));
#ifdef LUNAR_HAS_IPV6
			if (namefamily == AF_INET6 && ndx > -1) {
				
				if (TARGET_IS_IPV6(&namedb[ndx].name))
					myaddr = namedb[ndx].name;
				else {
					memset(&myaddr, 0, sizeof(struct lunartarget_s));
					memset(&myaddr, 0xfd, 1); // MANET scope
				}
				
			// With IPv6, we copy the target address in "myaddr"
			// In order to know what we are supposed to look for
			// as with IPv6 a node can have multiple addresses,
			// need address with similar scope (global, local)
			}
#endif
			NETBOX_Instance netbox_getL3addr(&myaddr,
						lunardev, namefamily);

			if (TARGET_IS_IPV6(&myaddr) && IN6_IS_ADDR_UNSPECIFIED(&myaddr.id.ipv6)) {
				DEBUG(DBG_INFO, printk("start_rreq: unspecified IPv6 source addr !\n"));
				return;
			}

			ups->is_bidir = 1;
			ups->backsel = peers[peerndx].path->backsel;

			send_rreq(radius, series, &ups->target,
				  ups->caller, ups->backsel,
				  &myhit,
				  myaddr.family ? &myaddr : 0,
#ifdef LUNAR_HAS_DNS
				  &myname,
#else
				  0,
#endif
				  0);
		} else
			send_rreq(radius, series, &ups->target,
				  ups->caller, 0, &myhit, 0, 0, 0);

		tout = radius * conf_timeout_per_ring;
	} else { // set up broadcast tree
		struct lunartarget_s *u;
				
		DEBUG(DBG_INFO, printk("BROADCAST !\n"));

		bps = (struct bcast_pending_s *)kmalloc(sizeof(struct bcast_pending_s), GFP_ATOMIC);
		if (!bps) {
			peers[peerndx].path->used = 0;
			return;
		}
		memset(bps, 0, sizeof(struct bcast_pending_s));
		bps->ndx = peerndx;
		bps->caller = mkHandlerPointer(
			NS_CLASS_addr bcast_rep_handler, bps);
		bps->fwd_sel = SAPF_Instance sapf_getUnusedSelector(mysapfhandle, SELCLASS_REMOTE);
		// bps->ia = peers[peerndx].ia;
		// bps->target.family = AF_INET;
		// bps->target.addr.ipv4 = peers[peerndx].name.addr.ipv4;
		u = namedb_getname(peerndx, AF_INET, 0);
		if (!u)
			u = namedb_getname(peerndx, AF_INET6, 0);// maybe should also check scope
		if (u) {
			bps->target = *u;
			send_rreq(radius, series, &bps->target, bps->caller,
				  0, 0, 0, 0, bps->fwd_sel);
		} else
		  DEBUG(DBG_INFO, printk("BROADCAST -- no target?\n"));
		tout = LUNAR_BRCAST_DELAY;
/*
		if (bc_len)
			tout = LUNAR_BRCAST_DELAY;
		else
			tout = conf_fri;
*/
// temporary:
		bc_sel = mkNetPointer(0, bps->fwd_sel);
	}

	NETBOX_Instance netbox_newtimer(&(peers[peerndx].path->timer),
					_VOID_VOID_ path_timeout,
					peers + peerndx, tout);
}


void NS_CLASS peer_timeout(void *vp)
{
	struct peer_s *n=(struct peer_s *)vp;
	int nndx = -1;

	DEBUG(DBG_PROC, printk("peer_timeout %d\n", n - peers));

	if (n->path) // restart timer
		NETBOX_Instance netbox_newtimer(&(n->timer),
						_VOID_VOID_ peer_timeout,
						vp, LUNAR_DNS_TIMEOUT);
	else { // erase entry
		
		nndx = namedb_get(n-peers, AF_NAME, 0);
		
		if (nndx >= 0) {
		
			if ( namedb[nndx].expires > (n->timer.expires + HZ) ) {
				
				/* struct lunartarget_s a;
				int i = 0; */
				
				DEBUG(DBG_PROC, printk("--> Lifetime of NAME entry is greater !\n"));
				NETBOX_Instance netbox_newtimer(&(n->timer),
					_VOID_VOID_ peer_timeout, vp,
					((namedb[nndx].expires-n->timer.expires)/HZ+1)*1000000);
				
				/* Private address maps version of LUNAR:
				
				--> binding lives as long as peer (to avoid a peer from changing
				IP addresses if app is standby for a long time
				
				while( i > -1 ) {
					i = namedb_getallipnames(n-peers, i, &a);
					if (i > -1) {
						DEBUG_TARGET(DBG_LUNAR, "--> Removing binding for", &a);
						namedb[i-1].name.family = 0;
						namedb[i-1].peerndx = -1;
					}
					
					n->flags &= ~PEER_IS_FAKE;
				} */
				
				return;
			}
		}
		
		namedb_delpeer(n - peers);
		n->flags = 0;
	}
}


void NS_CLASS path_timeout(void *vp)
{
	struct peer_s *n=(struct peer_s *)vp;
	struct lunartarget_s *t = 0;
	int nndx;
	struct uc_pending_s *ups = 0;

#ifdef LUNAR_HAS_DNS
	t = namedb_getname(n - peers, AF_NAME, 0);
#endif
	if (!t)
		t = namedb_getname(n - peers, AF_HIT, 0);
#ifdef LUNAR_HAS_IPV6
	if (!t)
		t = namedb_getname(n - peers, AF_INET6, IPV6_ADDR_SCOPE_MANET);
	if (!t)
		t = namedb_getname(n - peers, AF_INET6, IPV6_ADDR_SCOPE_GLOBAL);
#endif
	if (!t)
		t = namedb_getname(n - peers, AF_INET, 0);
	
	if (t) {
		DEBUG_TARGET(DBG_PROC, "path_timeout target:", t);
	} else
		DEBUG(DBG_PROC, printk("path_timeout %d / unknown target?\n",
				       n - peers));

	ups = sapf_getCallbackAux(mysapfhandle, n->upssel);

	if (!t && ups && ups->ndx > -1) {
		t = &namedb[ups->ndx].name;
		DEBUG(DBG_PROC, printk("--> using ups\n"));
	}
		
	if (!n->path || n->path->used == 0) {
		return;
	}
	DEBUG(DBG_PROC, printk("path_timeout (used=%d)\n", n->path->used));

	nndx = namedb_find(t);
	
	if (n->path->used == -1) {
		if (SAPF_Instance sapf_testandreset_active(mysapfhandle,
							   n->path->fwdsel)) {
			DEBUG(DBG_PROC, printk("restarting rreq because of activity\n"));
			if (n->path->flags & PATH_BCTREE)
				n->path->used = conf_max_ring + conf_max_retries;
			else
				n->path->used = MAX(n->path->hops, conf_max_ring);
			
			start_rreq(n - peers, nndx, 0, 0, t ? t->family : AF_INET);
		} else {
			DEBUG(DBG_PROC, printk("nonactive peer timeout\n"));

			SAPF_Instance sapf_deleteEntry(mysapfhandle,
						       n->path->fwdsel);
			if (n->path->flags & PATH_BCTREE) {
				DEBUG(DBG_PROC, printk("  bc_selp %p\n",
						       bc_selp));
				bc_sel = 0;
				if (bc_selp) {
					*bc_selp = 0;
					bc_selp = 0;
				}
			} else {
				struct lunartarget_s a;
				int i = 0;

				while( i > -1 ) {
					i = namedb_getallipnames(n-peers, i, &a);
					if (i > -1)
						NETBOX_Instance netbox_neigh_unmap(lunardev, &a);
				}
			}
			goto clean_up;
		}
		return;
	}
	
	if (n->path && (n->path->flags & PATH_BCTREE)) { // PEER_IS_BCAST) {

		DEBUG(DBG_INFO, printk("bc_delayed_start\n"));

		n->path->used = -1;
		n->path->hops = conf_max_ring - 1;
		n->path->fwdsel = bc_sel;
		n->path->flags |= PATH_IS_MAPPED;

		if (bc_len) {
			DEBUG(DBG_INFO, printk("bc_len\n"));
			if (bc_sel) {
				sapfpkt pkt;
				DEBUG(DBG_INFO, printk("bc_sel\n"));

				pkt.data = (char*) bc_buf;
				pkt.len = bc_len;
				SAPF_Instance sapf_demux(mysapfhandle, bc_sel, &pkt);
			}
			bc_len = 0;
		}
		if (bc_selp)
			*bc_selp = bc_sel;

		NETBOX_Instance netbox_newtimer(&(n->path->timer),
					_VOID_VOID_ path_timeout,
					n, conf_fri - LUNAR_BRCAST_DELAY);

		return;
	}

	if (n->path->used >= (conf_max_ring + conf_max_retries)) {
		int i=0, cnt;
		struct lunartarget_s a;
		// we give up
		DEBUG(DBG_LUNAR, printk("LUNAR path setup failed\n"));
		
#ifdef LUNAR_HAS_DNS
		if (ups) {
			if (ups->udp_port && ups->trans_id && ups->ndx >= 0) {
				DEBUG(DBG_INFO, printk("  DNS no such name\n"));
		
				send_dns_nosuchname(&namedb[ups->ndx].name,
					ups->udp_port, ups->trans_id, ups->qnamefamily);
			}
		}
#endif
		while( i > -1 ) {
			i = namedb_getallipnames(n-peers, i, &a);
			if (i > -1)
				NETBOX_Instance netbox_neigh_unmap(lunardev, &a);
		}
clean_up:
		memset(&n->upssel, 0, sizeof(sapfsel_t));
		peer_delpath(n);
		for (i = 0, cnt = 0; i < LUNAR_MAX_NAMES; i++) {
			if (namedb[i].name.family &&
			    namedb[i].peerndx == (n-peers))
				cnt++;
		}
		if (cnt <= 1) {// entry with no binding, therefore erase
			namedb_delpeer(n-peers);
			n->flags = 0;
		}
		return;
	}

	if (n->path->used < conf_max_ring)
		// 1-hop failed, so try immediately max hops
		n->path->used = (short) conf_max_ring;
	else    // count up to conf_max_retries
		n->path->used ++;
		
	if (ups) {
	
		start_rreq(n-peers, ups->ndx, ups->udp_port, 
				ups->trans_id, ups->qnamefamily);
	}
	else
		start_rreq(n - peers, nndx, 0, 0, t ? t->family : AF_INET);
}


// ----------------------------------------------------------------------

int NS_CLASS lunar_do_ndisc(struct lunartarget_s *target,
			    int udp_port, int trans_id, char namefamily)
{
	int peerndx, ndxp, ndx = -1;
	struct path_s *p;
	struct lunartarget_s myaddr;
	
	DEBUG_TARGET(DBG_PROC, "LUNAR: do_ndisc", target);

	peerndx = peer_getindex(target);
	
	// added for IP-less LUNAR
	
	if ( !TARGET_IS_NAME(target) && 
	     (target->id.ipv4.s_addr != htonl(LUNAR_IP_SUBNET + (unsigned char)(0xfe) )) &&
	     !namedb_getname( peerndx, AF_HIT, 0) )
	{
		// we explicitely forbid RREQ based on IP addresses
		// except if peer binding is known (we know its HIT)
		// TODO: send ICMP "no route to host"
		
		return -1;
	}
	
#ifdef LUNAR_HAS_IPV6
	if (TARGET_IS_IPV6(target) || namefamily == AF_INET6) {
	
		if( !IamIPv6Ready() ) {
			DEBUG(DBG_INFO, printk("  I'm not IPv6 enabled !\n"));
			
			if( udp_port && trans_id )
				send_dns_nosuchname(target, udp_port, trans_id, namefamily);
			return 0;
		}
	}
	
	if (TARGET_IS_IPV6(target)) {	
		
		memcpy(&myaddr, target, sizeof(struct lunartarget_s));
		NETBOX_Instance netbox_getL3addr(&myaddr, lunardev, AF_INET6);
		
		if (TARGET_EQUAL(target, &myaddr)) {
			DEBUG(DBG_LUNAR, printk("LUNAR: local lookup (DAD ?) --> skip RREQ\n"));
			return 0;
		}
	}
#endif	
		
#ifdef LUNAR_HAS_DNS
	NETBOX_Instance netbox_getL3addr(&myaddr, lunardev, AF_NAME);
	if (TARGET_EQUAL(target, &myaddr)) {
		DEBUG(DBG_LUNAR, printk("LUNAR: looking for myself ? --> skip RREQ\n"));
		memset( &myaddr, 0, sizeof(struct lunartarget_s));
		memset( (char*)&myaddr + 15, 0x01, 1); // IPv6 localhost ::1
		send_dns_reply(target, &myaddr, udp_port, trans_id, NOT_PTR);
		return 0;
	}
#endif

	if (peerndx < 0) {
		peerndx = peer_getnew();
		if (peerndx < 0)
			return 0;
		peer_set(peerndx, target);
	}
	
	ndx = namedb_find(target); // get namedb index of target
	
	if (ndx < 0) {
		DEBUG(DBG_PROC, printk("  --> failed to find namedb entry\n"));
		return 0;
	}
	
#ifdef LUNAR_HAS_DNS
	if (TARGET_IS_NAME(target)) {
		struct lunartarget_s *addr = 0;
		if( namefamily == AF_INET )
			addr = namedb_getname(peerndx, namefamily, 0);
#ifdef LUNAR_HAS_IPV6
		else {
			addr = namedb_getname(peerndx, namefamily, IPV6_ADDR_SCOPE_MANET);
			
			if (!addr)
				addr = namedb_getname(peerndx, namefamily, IPV6_ADDR_SCOPE_GLOBAL);
		}
#endif
		if (addr) {
			send_dns_reply(target, addr,
				       udp_port, trans_id, NOT_PTR);
			return 0;
		}
		
		goto skip;	// path might be up with wrong address family
				// TODO: use existing path for RREQ ?
	}
#endif
	if (peers[peerndx].path) {
		if (peers[peerndx].path->flags & PATH_IS_MAPPED) {
			lunar_neigh_map(peerndx);
			return 0;
		}
		// resolution is pending
		return -1;
	}

skip:
	ndxp = path_getnew();
	if (ndxp < 0)
		return -1;
	p = paths + ndxp;
	p->used = (short) conf_min_ring;
	p->backsel = SAPF_Instance sapf_getUnusedSelector(mysapfhandle,
							  SELCLASS_REMOTE);
	p->flags = PATH_IS_BIDIR;

	peers[peerndx].path = p;

	start_rreq(peerndx, ndx, udp_port, trans_id, namefamily);

	return -1;
}

 
void NS_CLASS lunar_do_bc_setup(sapfsel_t *sel, sapfpkt *pkt)
{
	struct lunartarget_s target;
	int peerndx;

	DEBUG(DBG_LUNAR, printk("lunar_do_bc_setup called\n"));

	target.family = AF_INET;
	target.id.ipv4.s_addr = htonl(LUNAR_IP_BCAST);

	peerndx = peer_getindex(&target);
	if (peerndx >= 0) {
		if (peers[peerndx].path &&
		    ((peers[peerndx].path->flags & PATH_IS_MAPPED) ||
		     (peers[peerndx].path->used > 0)))
			return;
	} else {
		peerndx = peer_getnew();
		if (peerndx < 0)
			return;
		peer_set(peerndx, &target);
	}

	if (!peers[peerndx].path) {
		int pndx = path_getnew();
		if (pndx < 0)
			return;
		peers[peerndx].path = paths + pndx;
	}
	DEBUG(DBG_PROC, printk("starting broadcast tree setup\n"));

	memcpy(bc_buf, pkt->data, pkt->len);
	bc_len = pkt->len;
	bc_sel = 0;
	bc_selp = sel;
	
	peers[peerndx].path->used = conf_max_ring + conf_max_retries;
	peers[peerndx].path->backsel = SAPF_Instance
		sapf_getUnusedSelector(mysapfhandle, SELCLASS_REMOTE);
	peers[peerndx].path->flags |= PATH_BCTREE;

	start_rreq(peerndx, -1, 0, 0, AF_INET);
}


// eof
