/* lunar_dhcp.c
 *
 *   LUNAR - Lightweight Underlay Network Ad hoc Routing
 *
 *   DHCP interception and address self-config
 *
 * Copyright (C) 2003-2005 <christian.tschudin@unibas.ch>
 *
 *   2005-01-26 Christophe Jelger: IPv6 stuff
 *   2004-11-13 Per Svensson: fixed some warnings about
 *                  incompatible pointer type
 *   2004-05-15 now interoperates with windows DHCP client
 *   2003-11-20 initial version of dhcp intercept
 *
 * This file is distributed under the terms of the GNU general Public
 * License (GPL), see the file COPYING
 */

#ifdef LUNAR_HAS_DHCP

static int asearch_rrep_handler(sapfpkt *p, void *dummy);
static void asearch_timeout(void *dummy);

struct udp_pseudo_hdr {
	uint32_t src, dst;
	unsigned char zero:8, proto:8;
	uint16_t len:16;
};

struct bootphdr {
	unsigned char opcode;
	unsigned char hw;
	unsigned char hwlength;
	unsigned char hopcount;
	uint32_t xid;
	uint16_t secs;
	uint16_t flags;
	uint32_t ciaddr, yiaddr, server_ip, bootp_gw_ip;
	unsigned char hwaddr[16];
	unsigned char servername[64];
	unsigned char bootfile[128];
} ;

#define DHCP_DISCOVER 1
#define DHCP_OFFER    2
#define DHCP_REQUEST  3
#define DHCP_ACK      5
#define DHCP_NACK     6

static struct addr_search_s {
	sapfsel_t callback;		// local handler to remove
	struct netbox_timer_s timer;
	unsigned char attempts;		// decrementing, 0=inactive
	unsigned char verify;		// retries
	uint32_t ip;			// addr currently probed
	unsigned char done[32];		// bitset of visited addresses
	unsigned char bootp[sizeof(struct bootphdr)];
	char is_request;
} asearch;



// ----------------------------------------------------------------------

static void send_dhcp(char dhcptype, uint32_t addr)
{
	unsigned char buf[512];
	int len, len2;
	struct iphdr *ih = (struct iphdr*) buf;
	struct udphdr *uh = (struct udphdr*) (buf + sizeof(struct iphdr));
	struct bootphdr *bh = (struct bootphdr*) (buf + sizeof(struct iphdr) + sizeof(struct udphdr));
	struct udp_pseudo_hdr *uph;

        DEBUG(9, printk("building DHCP reply: %d %s\n", dhcptype, lunar_ip2str(addr)));
	memset(buf, 0, sizeof(buf));

	ih->ihl = 5;
	ih->version = 4;
	ih->ttl = 2;
	ih->protocol = 17;
	ih->saddr = htonl(LUNAR_IP_SUBNET | 0xfe);
	ih->daddr = 0xffffffff;
	uh->source = htons(67);
	uh->dest = htons(68);
	memcpy(bh, asearch.bootp, sizeof(asearch.bootp));

	bh->opcode = 2; // BOOTP reply
	bh->yiaddr = addr;
//	bh->server_ip = 0;
	bh->server_ip = htonl(LUNAR_IP_SUBNET | 0xfe);
	memset(bh->servername, 0, sizeof(bh->servername));

	// magic
	len = (unsigned char*)(bh + 1) - buf;
	memcpy(buf+len, "\x63\x82\x53\x63", 4);
	len += 4;

	// DHCP message type
	memcpy(buf+len, "\065\001", 2);
	len += 2; buf[len++] = dhcptype;

	if (dhcptype != DHCP_NACK) {
		// subnet mask
		memcpy(buf + len, "\001\004\xff\xff\xff\x00", 6);
		len += 6;

		// Lease time
		memcpy(buf + len, "\063\004\x00\x00\x20\x00", 6);
		len += 6;

#ifdef LUNAR_HAS_DNS
		// domain name
		memcpy(buf + len, "\017\011", 2);
		memcpy(buf + len + 2, domain_name, strlen(domain_name));
		len += (2 + strlen(domain_name));
#endif

		// faked lunar name server
		memcpy(buf + len, "\006\004", 2);
		*(uint32_t *)(buf + len + 2) = htonl(LUNAR_IP_NAMESERVER);
		len += 6;

/*
		// dgw
		memcpy(buf + len, "\003\004\x01\x02\x03\x06", 6);
		len += 6;
*/
	}
	// server identifier option
	memcpy(buf + len, "\066\004", 2);
	*(uint32_t *)(buf + len + 2) = htonl(LUNAR_IP_SUBNET | 0xfe);
	len += 6;

	buf[len] = 0xff; // eolist
	len++;

	ih->tot_len = htons((__u16)len);
	uh->len = htons(len - sizeof(struct iphdr));

	// build the UDP pseudoheader (appended for doing the chksum)
	len2 = (len + 3) & ~0x03; // dword align
	uph = (struct udp_pseudo_hdr *) (buf + len2);
	uph->src = ih->saddr;
	uph->dst = ih->daddr;
	uph->proto = ih->protocol;
	uph->len = uh->len;
	len2 += sizeof(*uph);

	uh->check = ip_fast_csum((char*) uh, (len2 - sizeof(struct iphdr))/4);
	if (!uh->check)
		uh->check = 0xffff;
	ih->check = ip_fast_csum(buf, ih->ihl);
	// printk("- ip chksum 0x%04x, ip dst 0x%08x, ip len %d\n", ih->check, ih->daddr, len);

	netbox_l3deliver(lunardev->ifindex, buf, len);
}


static void asearch_send_probe(uint32_t addr)
{
	sapfsel_t series;
	struct lunartarget_s target;
	
	target.family = AF_INET;
	target.id.ipv4.s_addr = addr;

        DEBUG(9, printk("asearch_send_probe %08x\n", ntohl(addr)));

	asearch.ip = addr;
	asearch.callback = mkHandlerPointer(
			(sapfcallback_t) asearch_rrep_handler, 0);

	series = sapf_getUnusedSelector(mysapfhandle, SELCLASS_LOCAL);
	sapf_addREDIR(mysapfhandle, series, 0);
	send_rreq(conf_max_ring, series, &target,
		  asearch.callback, 0, 0, 0, 0, 0);
	netbox_newtimer(&asearch.timer, asearch_timeout, 0, LUNAR_IP_PROBE_RETRY);
}


static uint32_t randomLUNARaddr(void)
{
	unsigned char h, i;
	uint32_t a;

	for (i = 255; i > 0; i--) {
		get_random_bytes(&h, 1);
		a = LUNAR_IP_SUBNET | h;
		if (!test_bit(h, (void*)asearch.done))
			break;
	}
	return htonl(a);
}


static int asearch_rrep_handler(sapfpkt *p, void *dummy)
{
	struct xrp_head_s *x = (struct xrp_head_s*) p->data;

        DEBUG(9, printk("asearch_rrep_handler (%d bytes)\n", p->len));

	if (p->len < sizeof(struct xrp_head_s) ||
					    x->cmd != XRP_CMD_LUNAR_RREP)
		return 0;

	sapf_deleteEntry(mysapfhandle, asearch.callback);
	netbox_deltimer(&asearch.timer);

	if (asearch.is_request) {
		asearch.attempts = 0;
		send_dhcp(DHCP_NACK, asearch.ip);
	} else {
		set_bit(ntohl(asearch.ip) & 0xff, (void*)asearch.done);
		asearch.attempts--;
		if (asearch.attempts > 0) {
			asearch.ip = randomLUNARaddr();
			asearch_send_probe(asearch.ip);
		}
		// else we give up - we send no reply to the client's DISCOVER
	}

	return 0;
}


static void asearch_timeout(void *dummy)
{
        DEBUG(9, printk("asearch_timeout (attempts=%d, verify=%d)\n",
			asearch.attempts, asearch.verify));

	sapf_deleteEntry(mysapfhandle, asearch.callback);

	asearch.verify++;
	if (asearch.verify > LUNAR_IP_VERIFY_MAX) { // address seems to be clean
		asearch.attempts = 0;
		if (asearch.is_request) {
			// ... would have to also collect gateways
			send_dhcp(DHCP_ACK, asearch.ip);
		} else
			send_dhcp(DHCP_OFFER, asearch.ip);
	} else
		asearch_send_probe(asearch.ip);
}


static void lunar_do_dhcp(sapfpkt *pkt)
{
	unsigned char *cp;
	struct iphdr *ih = (struct iphdr*)pkt->data;
	struct udphdr *uh;
	struct bootphdr *bh;
	uint32_t rq = 0;

	DEBUG(10, printk("lunar DHCP\n"));

	uh = (struct udphdr*) (ih + 1); // XXX should look for IP options
	bh = (struct bootphdr*) ( uh + 1);
/*
	if (pkt->len <= (unsigned char*) (bh + 1) - (unsigned char*)pkt->data) {
		DEBUG(10, printk("  too short\n"));
		return;
	}
*/

	// verify that it's a DHCP packet:
	cp = (unsigned char*) (bh + 1) + 4;
	while ((cp - (unsigned char*)pkt->data) < pkt->len && *cp != 0xff) {
		if (!*cp) { cp++; continue; }
		if (*cp == 53) { cp += 2; break; }
		cp += 2 + cp[1];
	}
	if (cp - (unsigned char*)pkt->data >= pkt->len || *cp == 0xff) {
		DEBUG(10, printk("  did not find DHCP option %d %d %d\n",
				 (char*)(bh+1) -(char*)pkt->data,
				 (char*)(uh+1) -(char*)pkt->data,
				 (char*)(ih+1) -(char*)pkt->data));
		return;
	}
	if ( *cp != DHCP_DISCOVER && *cp != DHCP_REQUEST) {
		DEBUG(10, printk("  DHCP refused: option type was %d\n", *cp));
		return;
	}

	if (asearch.attempts) {
		netbox_deltimer(&asearch.timer);
		sapf_deleteEntry(mysapfhandle, asearch.callback);
		asearch.attempts = 0;
	}
	asearch.verify = 0;

	asearch.is_request = *cp == DHCP_REQUEST;

	memcpy(asearch.bootp, bh, sizeof(struct bootphdr));
	asearch.attempts = 10;

	memset(asearch.done, 0, sizeof(asearch.done));
	set_bit(0, (void*)asearch.done);
	set_bit(254, (void*)asearch.done);
	set_bit(255, (void*)asearch.done);

	cp = (unsigned char*) (bh + 1) + 4;
	while ((cp - (unsigned char*)pkt->data) < pkt->len && *cp != 0xff) {
		if (!*cp) { cp++; continue; }
		if (*cp == 50) { cp += 2; rq = *(uint32_t*)cp; break; }
		cp += 2 + cp[1];
	}

	DEBUG(9, printk("%s : %s\n", asearch.is_request ? 
			"DHCP_REQUEST" : "DHCP_DISCOVER", lunar_ip2str(rq)));

	if (!rq && asearch.is_request) {
		struct lunartarget_s a;
		netbox_getL3addr(&a, lunardev, AF_INET);
		if (a.id.ipv4.s_addr)
			rq = a.id.ipv4.s_addr;
	}

	if ((LUNAR_IP_NETMASK & ntohl(rq)) != LUNAR_IP_SUBNET) {
		if (asearch.is_request) {
			send_dhcp(DHCP_NACK, rq);
			return;
		}
		rq = 0;
	}
	
	asearch.ip = rq ? rq : randomLUNARaddr();
	
	/* Private address maps version: no need to test address
	asearch_send_probe(asearch.ip); 
	*/
	
	if (asearch.is_request)
		send_dhcp(DHCP_ACK, asearch.ip);
	else
		send_dhcp(DHCP_OFFER, asearch.ip);
}

#endif // LUNAR_HAS_DHCP

// eof
