/* lunar_dns.c
 *
 *   LUNAR - Lightweight Underlay Network Ad hoc Routing
 *
 *   Lunar logic
 *
 * Copyright (C) 2005 <Christophe.Jelger@unibas.ch>
 *
 * This file is distributed under the terms of the GNU general Public
 * License (GPL), see the file COPYING
 *
 *  2005-07-21 cj: found big bug (related to ip6.arpa) in fill_addr_to_dns
 *  2005-07-06 cj: added send_dns_nosuchname() --> DNS error msg
 *                 added PTR lookup for IPv6 localhost ::1
 *  2005-05-26 cj: added old IPv6 PTR style
 *  2005-02-18 cft: code adaption
 *  2005-01-26 cj: first version
 */

#ifdef LUNAR_HAS_DNS

#define DNS_REQ_A           1
#define DNS_REQ_AAAA       28
#define DNS_REQ_PTR        12

#define PTR_STD_STYLE      0
#define PTR_OLD_STYLE      1
#define NOT_PTR            2

// and PTR queries can be very long with IPv6 !
#define MAX_QUERY_LEN     128

static char tmpname[MAX_QUERY_LEN];
static char tmpaddr[MAX_QUERY_LEN];

struct dns_header {
	uint16_t trans_id;
	uint16_t flags;
	uint16_t qdcount;
	uint16_t ancount;
	uint16_t nscount;
	uint16_t arcount;
};

struct udp_pseudo_hdr2 {
	uint32_t src, dst;
	unsigned char zero, proto;
	uint16_t len;
};

char *dotted_domain = LUNAR_DOTTED_DOMAIN_NAME;

char *dsdnarray[MAX_DSDN];

// ----------------------------------------------------------------------

int NS_CLASS fill_name_to_dns(char *dns, char *name)
{
	char *cp = dns;
	int i;

	// reconstruct DNS-like name from human dotted name
	// e.g. www.renater.fr --> 03www07renater02fr00
	while (*name) {
		for (i = 0; name[i] && name[i] != '.'; i++);
		*cp = i;
		if (i) {
			cp++;
			strncpy(cp, name, i);
			cp += i;
		}
		name += i;
		if (*name) // skip the dot
			name++;
	}
	*cp++ = 0;		// set \0 at end of name

	return cp - dns;

}

#ifdef LUNAR_HAS_IPV6
int NS_CLASS fill_addr_to_dns_oldstyle(char *dns, struct lunartarget_s *addr)
{
	unsigned char *cp, *ap;
	uint16_t *sp;

	ap = (unsigned char*) &(addr->id.ipv6);
	sp = (uint16_t *) dns;
	*sp = htons(0x4180); // old PTR style
	cp = (dns+2);
		
	memcpy( cp, ap, sizeof(struct in6_addr));
	cp += sizeof(struct in6_addr);
	
	*cp = 3;
	strcpy(cp+1, "ip6");
	cp += *cp + 1;

	*cp = 4;
	strcpy(cp+1, "arpa");
	cp += *cp + 1;

	*cp++ = 0; // end marker

	return cp - (unsigned char*) dns;
}
#endif

int NS_CLASS fill_addr_to_dns(char *dns, struct lunartarget_s *addr)
{
	unsigned char *cp, *ap;
	char alen;
	char *fmt;

	if (addr->family == AF_INET) {
		alen = sizeof(struct in_addr);
		ap = (unsigned char*) &(addr->id.ipv4);
		fmt = "%d";
		
		ap += alen - 1;
		for (cp = dns; alen > 0; alen--, ap--) {
			*cp = sprintf(cp+1, fmt, *ap);
			cp += *cp + 1;
		}
		
		*cp = 7;
		strcpy(cp+1, "in-addr");
		cp += *cp + 1;
		
	} else {
#ifdef LUNAR_HAS_IPV6
		alen = sizeof(struct in6_addr);
		ap = (unsigned char*) &(addr->id.ipv6);
		fmt = "%x";
		
		ap += alen - 1;
		for (cp = dns; alen > 0; alen--, ap--) {
			*cp = sprintf(cp+1, fmt, *ap & 0x0f);
			cp += *cp + 1;
			*cp = sprintf(cp+1, fmt, (*ap & 0xf0)>>4);
			cp += *cp + 1;
		}
		
		*cp = 3;
		strcpy(cp+1, "ip6");
		cp += *cp + 1;
#else
		DEBUG(0, printk("fill_addr_to_dns assert failed\n"));
		return 0;
#endif
	}

	*cp = 4;
	strcpy(cp+1, "arpa");
	cp += *cp + 1;

	*cp++ = 0; // end marker

	return cp - (unsigned char*) dns;
}

int NS_CLASS fill_dns_to_addr(struct lunartarget_s *addr, char *dns)
{
	int i;
	uint32_t ip;
	unsigned char *cp = dns;

	// extract v4 addr from DNS PTR format
	for (i = 0, ip = 0; i < 4; i++) {
		int v;
		for (v = 0; *cp; cp++) {
			if (*cp == '.') {
				cp++;
				break;
			}
			v = 10*v + *cp - '0';
		}
		ip = ((v & 0xff) << 24) | (ip >> 8);
	}
	addr->id.ipv4.s_addr = htonl(ip);
	addr->family = AF_INET;

	return cp - (unsigned char*) dns;
}

void NS_CLASS send_dns_reply(struct lunartarget_s *query,
			     struct lunartarget_s *reply,
			     int udp_port, int query_id, int ptrstyle)
{
	// DNS/IPv4 replies (and queries) only
	unsigned char buf[512];
	struct iphdr *iph        = (struct iphdr *) buf;
	struct udphdr *udph      = (struct udphdr *) (iph + 1);
	struct dns_header *dnsh  = (struct dns_header *) (udph + 1);
	unsigned char *cp        = (unsigned char *) (dnsh + 1);
	struct udp_pseudo_hdr2 *uph;
	uint16_t *sp;
	int len = sizeof(struct iphdr) + sizeof(struct udphdr);
	int len2;
	struct lunartarget_s myself;
	int qtype, alen, qlen;
	void *qp, *ap; // pointers to query and answer data

	NETBOX_Instance netbox_getL3addr(&myself, lunardev, AF_INET);

	memset(buf, 0, sizeof(buf));

	// IPv4 header of reply
	iph->ihl = 5;
	iph->version = 4;
	iph->ttl = 2;
	iph->protocol = IPPROTO_UDP;
	iph->saddr = htonl(LUNAR_IP_SUBNET | 0xfe);
	iph->daddr = myself.id.ipv4.s_addr;

	// UDP reply is DNS to original port
	udph->source = htons(53);
	udph->dest = udp_port;

	// DNS reply message header
	dnsh->trans_id = query_id;
	dnsh->flags = htons(0x8180);	// standard reply
	dnsh->qdcount = htons(1);
	dnsh->ancount = htons(1);
	dnsh->nscount = 0;
	dnsh->arcount = 0;

	switch (query->family) {

	case AF_NAME:
		if (reply->family == AF_INET) {
			qtype = DNS_REQ_A;
			qp = tmpname;
			qlen = fill_name_to_dns(tmpname, query->id.name);
			ap = &(reply->id.ipv4);
			alen = sizeof(struct in_addr);
		}
#ifdef LUNAR_HAS_IPV6
		else {
			qtype = DNS_REQ_AAAA;
			qp = tmpname;
			qlen = fill_name_to_dns(tmpname, query->id.name);
			ap = &(reply->id.ipv6);
			alen = sizeof(struct in6_addr);
		}
#else
		else return;
#endif		
		break;
	case AF_INET:
		qtype = DNS_REQ_PTR;
		qp = tmpaddr;
		qlen = fill_addr_to_dns(tmpaddr, query);
		ap = tmpname;
		alen = fill_name_to_dns(tmpname, reply->id.name);
		break;
#ifdef LUNAR_HAS_IPV6
	case AF_INET6:
		qtype = DNS_REQ_PTR;
		qp = tmpaddr;
		if (ptrstyle == PTR_OLD_STYLE )
			qlen = fill_addr_to_dns_oldstyle(tmpaddr, query);
		else
			qlen = fill_addr_to_dns(tmpaddr, query);
		ap = tmpname;
		alen = fill_name_to_dns(tmpname, reply->id.name);
		break;
#endif
	default:
		DEBUG(0, printk("send_dns_reply: assert failed\n"));
		return;
	}
	// append query part
	memcpy(cp, qp, qlen);
	cp += qlen;
	// no padding? (for 16bit alignment?)
	sp = (uint16_t *) cp;
	*sp++ = htons(qtype);
	*sp++ = htons(1);
	*sp++ = htons(0xc00c); // compression
	*sp++ = htons(qtype);
	*sp++ = htons(1);	// set class (1 for Internet)
	*(uint32_t *)sp = htonl(600);	// TTL : we use 10mn
	sp += 2;
	// append response part
	*sp++ = htons(alen);
	memcpy(sp, ap, alen);
	len += (unsigned char*) sp - (unsigned char*) dnsh + alen;

	iph->tot_len = htons(len);
	udph->len = htons(len - sizeof(struct iphdr));

	// build the UDP pseudoheader (appended for doing the chksum)
	len2 = (len + 3) & ~0x03;	// dword align
	uph = (struct udp_pseudo_hdr2 *) (buf + len2);
	uph->src = iph->saddr;
	uph->dst = iph->daddr;
	uph->proto = iph->protocol;
	uph->len = udph->len;
	len2 += sizeof(*uph);

	udph->check =
	    ip_fast_csum((char *) udph, (len2 - sizeof(struct iphdr)) / 4);
	if (!udph->check)
		udph->check = 0xffff;
	iph->check = ip_fast_csum(buf, iph->ihl);

	netbox_l3deliver(lunardev->ifindex, buf, len);
}

void NS_CLASS send_dns_nosuchname(struct lunartarget_s *query, 
				int udp_port, int query_id, char family)
{
	// DNS/IPv4 replies (and queries) only
	unsigned char buf[512];
	struct iphdr *iph        = (struct iphdr *) buf;
	struct udphdr *udph      = (struct udphdr *) (iph + 1);
	struct dns_header *dnsh  = (struct dns_header *) (udph + 1);
	unsigned char *cp        = (unsigned char *) (dnsh + 1);
	struct udp_pseudo_hdr2 *uph;
	int len = sizeof(struct iphdr) + sizeof(struct udphdr);
	int len2;
	uint16_t *sp;
	struct lunartarget_s myself;
	int qtype, qlen;
	void *qp; // pointers to query and answer data

	NETBOX_Instance netbox_getL3addr(&myself, lunardev, AF_INET);

	memset(buf, 0, sizeof(buf));

	// IPv4 header of reply
	iph->ihl = 5;
	iph->version = 4;
	iph->ttl = 2;
	iph->protocol = IPPROTO_UDP;
	iph->saddr = htonl(LUNAR_IP_SUBNET | 0xfe);
	iph->daddr = myself.id.ipv4.s_addr;

	// UDP reply is DNS to original port
	udph->source = htons(53);
	udph->dest = udp_port;

	// DNS reply message header
	dnsh->trans_id = query_id;
	dnsh->flags = htons(0x8583);	// "No such name"
	dnsh->qdcount = htons(1);
	dnsh->ancount = 0;
	dnsh->nscount = 0;
	dnsh->arcount = 0;

	switch (family) {

	case AF_INET:
		qtype = DNS_REQ_A;
		qp = tmpname;
		qlen = fill_name_to_dns(tmpname, query->id.name);
		break;
#ifdef LUNAR_HAS_IPV6
	case AF_INET6:
		qtype = DNS_REQ_AAAA;
		qp = tmpname;
		qlen = fill_name_to_dns(tmpname, query->id.name);
		break;
#endif		
	default:
		DEBUG(0, printk("send_dns_nosuchname: assert failed\n"));
		return;
	}
	// append query part
	memcpy(cp, qp, qlen);
	cp += qlen;
	sp = (uint16_t *) cp;
	*sp++ = htons(qtype);
	*sp++ = htons(1);
	len += (unsigned char*) sp - (unsigned char*) dnsh;
	
	iph->tot_len = htons(len);
	udph->len = htons(len - sizeof(struct iphdr));

	// build the UDP pseudoheader (appended for doing the chksum)
	len2 = (len + 3) & ~0x03;	// dword align
	uph = (struct udp_pseudo_hdr2 *) (buf + len2);
	uph->src = iph->saddr;
	uph->dst = iph->daddr;
	uph->proto = iph->protocol;
	uph->len = udph->len;
	len2 += sizeof(*uph);

	udph->check = ip_fast_csum((char *) udph, (len2 - sizeof(struct iphdr)) / 4);
	if (!udph->check)
		udph->check = 0xffff;
	iph->check = ip_fast_csum(buf, iph->ihl);

	netbox_l3deliver(lunardev->ifindex, buf, len);
}

void NS_CLASS lunar_dns_ptr4(char *cp, struct lunartarget_s *addr)
{
	int i;
	uint32_t ip;

	// extract v4 addr from DNS PTR format
	for (i = 0, ip = 0; i < 4; i++) {
		int v;
		for (v = 0; *cp; cp++) {
			if (*cp == '.') {
				cp++;
				break;
			}
			v = 10*v + *cp - '0';
		}
		ip = ((v & 0xff) << 24) | (ip >> 8);
	}		
	addr->id.ipv4.s_addr = htonl(ip);
	addr->family = AF_INET;
	DEBUG_TARGET(9, "(v4) -->", addr);
}


#ifdef LUNAR_HAS_IPV6

void NS_CLASS lunar_dns_ptr6(char *ptr, struct lunartarget_s *addr)
{
	int i;

	// v6 PTR are in DNS-format
	// --> extracts v6 addr from DNS PTR format

	// ptr++;	// skip len indicator

	for (i = 15; i >= 0; i--) {
		unsigned char c;

		if ((addr->id.ipv6.s6_addr[i] = *ptr - '0') > 10)
			addr->id.ipv6.s6_addr[i] -= 39;

		ptr += 2;	// next value

		if ((c = *ptr - '0') > 10)
			c -= 39;

		addr->id.ipv6.s6_addr[i] += c * 0x10;

		ptr += 2;	// next value
	}

	addr->family = AF_INET6;
	DEBUG_TARGET(9, "(v6) -->", addr);
}

#endif // LUNAR_HAS_IPV6


// ----------------------------------------------------------------------

int NS_CLASS lunar_do_dns(sapfpkt *pkt)
{
	struct iphdr *iph       = (struct iphdr *) (pkt->data);
	struct udphdr *udph     = (struct udphdr *) (iph + 1);
	struct dns_header *dnsh = (struct dns_header *) (udph + 1);
	unsigned char *cp = (unsigned char *) (dnsh + 1);
	uint16_t qtype;
	int ndx, len;
	struct lunartarget_s target;
	struct lunartarget_s *name;
	char tmp_name[MAX_QUERY_LEN + 1];
	char replynamefamily = AF_INET;
	int ptrstyle = PTR_STD_STYLE;

	if (dnsh->flags & 0xF800) {
		DEBUG(9, printk("name_lookup: non-standard query!\n"));
		return 1;
	}
	DEBUG(9, printk("name_lookup: ID=%d ", ntohs(dnsh->trans_id)));

	// check for deprecated PTR format 0x4180 for IPv6
	if (ntohs(*(uint16_t *)cp) == 0x4180) {
		DEBUG(10, printk("is in bit-string label format\n"));
#ifdef LUNAR_HAS_IPV6
		memcpy(&target.id.ipv6.s6_addr, cp+2, 16);
		target.family = AF_INET6;
		ndx = peer_getindex(&target);
		ptrstyle = PTR_OLD_STYLE;
#else
		ndx = -1;
#endif
		goto ptr_lookup;
	}

	len = 0;
	while (*cp) {
		int i = *cp;
		cp++;

		while (i--) {
			if (len < MAX_QUERY_LEN)
				tmp_name[len++] = *cp;
			cp++;
		}

		if (len < MAX_QUERY_LEN && *cp)
			tmp_name[len++] = '.';
	}
	tmp_name[len] = '\0';
	cp++;
	DEBUG(9, printk(" (%s %d)", tmp_name, cp - (unsigned char *) (dnsh + 1)));

	qtype = ntohs(* (uint16_t *) cp);
	cp += 2;
	DEBUG(9, printk(" TYPE=%d CLASS=%d\n", qtype,
			ntohs(* (uint16_t *) cp)));

	switch (qtype) {
	case DNS_REQ_PTR:
		DEBUG(9, printk("PTR request ! "));
		if (len <= 64) {
			lunar_dns_ptr4(tmp_name, &target);
#ifdef LUNAR_HAS_IPV6
		} else
			// an IPv6 PTR is always greater than 64
			lunar_dns_ptr6(tmp_name, &target);
			ndx = peer_getindex(&target);
#else
			ndx = peer_getindex(&target);
		} else
			ndx = -1;
#endif
ptr_lookup:
		if (ndx < 0) {
			struct lunartarget_s trgt;
#ifdef LUNAR_HAS_IPV6			
			memset( &trgt, 0, sizeof(struct lunartarget_s));
			memset( (char*)&trgt + 15, 0x01, 1); // IPv6 localhost ::1
			trgt.family = AF_INET6;
			
			if (TARGET_EQUAL(&target, &trgt)) {
				DEBUG(9, printk("name_lookup: localhost (::1)\n"));
				NETBOX_Instance netbox_getL3addr(&trgt, lunardev,
							 AF_NAME);
				send_dns_reply(&target, &trgt,
				       udph->source, dnsh->trans_id, ptrstyle);
				return 1;
			}
#endif
			
			NETBOX_Instance netbox_getL3addr(&trgt,
							 lunardev, target.family);
			if (!TARGET_EQUAL(&target, &trgt)) {
				DEBUG_TARGET(9, "name_lookup: unknown addr",
					     &target);
				return 1;
			}
			NETBOX_Instance netbox_getL3addr(&trgt, lunardev,
							 AF_NAME);
			DEBUG_TARGET(9, "name_lookup: name is", &trgt);
			send_dns_reply(&target, &trgt,
				       udph->source, dnsh->trans_id, ptrstyle);
			return 1;
		}
		DEBUG_TARGET(9, "name_lookup: known addr", &target);
		name = namedb_getname(ndx, AF_NAME, 0);
		if (name) {
			DEBUG_TARGET(9, "name_lookup: name is", name);
			send_dns_reply(&target, name,
				       udph->source, dnsh->trans_id, ptrstyle);
		}
		return 1;

#ifdef LUNAR_HAS_IPV6
	case DNS_REQ_AAAA:
		replynamefamily = AF_INET6;
		// we should keep this in mind when doing the search ...
#endif
	case DNS_REQ_A:
		target.family = AF_NAME;
		break;
	default:
		return 1;
	}

	//----- Check that A or AAAA name belongs to .net.lunar
	// else we will do a name request towards a gateway
	// This still needs to be defined

	strncpy(target.id.name, tmp_name, LUNAR_MAX_HOSTNAME_LEN + 1);

	if (strcmp(dotted_domain, tmp_name + len - strlen(dotted_domain))) {
		DEBUG(9, printk("NOT a net.lunar request !\n"));
		send_dns_nosuchname(&target, udph->source, dnsh->trans_id, replynamefamily);
		return 0;
	}

	/* remove uppercase */
	for ( cp=target.id.name; *cp != '.'; cp++ ) {
		if ( *cp >= 'A' && *cp <= 'Z' )
			*cp += 32;
	}

	lunar_do_ndisc(&target, udph->source, dnsh->trans_id, replynamefamily);

	return 1;
}


void NS_CLASS lunar_dns_server(sapfpkt *pkt, void *aux)
{
	DEBUG(1, printk("nameserver callback\n"));
}

int NS_CLASS lunar_parse_dsdnparams(char *params, char **array)
{
	int cnt = 0;
	
	if (!params || !array)
		return 0;
	
	while (params && cnt < MAX_DSDN) {
			
		array[cnt] = strsep(&params, ",");
		DEBUG(1, printk("array[%d] = %s\n", cnt, array[cnt]));
		if (!array[cnt])
			break;
		cnt++;
	}
	
	return cnt-1;
}

void NS_CLASS lunar_dns_init(void)
{
	int ndx;
	sapfsel_t dnsdeliver;
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,17)
	char *sysname = utsname()->nodename;
#else
	char *sysname = system_utsname.nodename;
#endif	

        dnsdeliver = SAPF_Instance sapf_getUnusedSelector(mysapfhandle,
							  SELCLASS_LOCAL);
        SAPF_Instance sapf_addCallback(mysapfhandle, dnsdeliver,
                                       lunar_dns_server, 0);
        SAPF_Instance sapf_setPermanent(mysapfhandle, dnsdeliver);

	ndx = peer_getnew();
	if (ndx >= 0) {
		struct lunartarget_s name_server;
		// Note that an ARP table entry will be automatically
		// added the first time a request is made for the
		// IP address "LUNAR_NAME_SERVER"
		name_server.id.ipv4.s_addr = htonl(LUNAR_IP_NAMESERVER);
		name_server.family = AF_INET;
		peer_set(ndx, &name_server);

		strcpy(name_server.id.name, "lunar-dns" LUNAR_DOTTED_DOMAIN_NAME);
		name_server.family = AF_NAME;
		peer_set(ndx, &name_server);

		peers[ndx].path = kmalloc(sizeof(struct path_s), GFP_ATOMIC);
		if (peers[ndx].path) {
			peers[ndx].path->fwdsel = dnsdeliver;
			peers[ndx].path->flags |= PATH_IS_MAPPED;
			peers[ndx].path->used = -1;
			peers[ndx].path->hops = 0;
			
		}
	}
	
	/* Dynamic Shortest Discriminating Name */
	/* Start with firstname.familyname */
	
	memset(dsdnarray, 0, MAX_DSDN);
	
	if (dsdnparams)
	{
		char *cp;
		
		/* remove uppercase */
		for ( cp = dsdnparams; *cp != '\0'; cp++ ) {
			if ( *cp >= 'A' && *cp <= 'Z' )
				*cp += 32;
		}
	}
	else
	{	strcpy(dsdname, sysname);
		dsdnvector = 0;
		return;
	}
	
	if (strlen(dsdnparams) == 0 )
	{
		strcpy(dsdname, sysname);
		dsdnvector = 0;
	}
	else
	{
		/* build initial LUNAR name */
		
		int cnt;
				
		strcpy(dsdnparamscopy, dsdnparams);
		cnt = lunar_parse_dsdnparams(dsdnparamscopy, dsdnarray);
		
		if (cnt == 0)
		{
			strcpy(dsdname, sysname);
			dsdnvector = 0;
		}
		else if (cnt >= 2)
		{
			strcpy(dsdname, dsdnarray[0]);
			strcpy(dsdname+strlen(dsdname), "-");
			strcpy(dsdname+strlen(dsdname), dsdnarray[1]);
			dsdnvector = 12;
		}
		else
		{
			strcpy(dsdname, dsdnarray[0]);
			dsdnvector = 1;
		}
	}
	
	/* start scheduling DSDN "reduce" function */
	
	NETBOX_Instance netbox_newtimer(&dsdntimer, 
			_VOID_VOID_ lunar_dsdn_reduce, 0, LUNAR_DSDN_TIMEOUT);
}


void NS_CLASS lunar_dns_fini(void)
{
        int ndx;
        struct lunartarget_s name_server;

	DEBUG(DBG_PROC, printk("LUNAR: dns_fini\n"));

        name_server.id.ipv4.s_addr = htonl(LUNAR_IP_NAMESERVER);
        name_server.family = AF_INET;

        // remove_peer();
        ndx = peer_getindex(&name_server);
        if (ndx >= 0)
		peer_del(peers + ndx);
}

#endif				// LUNAR_HAS_DNS

// eof
