/* lunar_ipv6.c
 *
 *   LUNAR - Lightweight Underlay Network Ad hoc Routing
 *
 *   IPv6 interception and address self-config
 *
 * Copyright (C) 2005 <christophe.jelger@unibas.ch>
 *
 *   2005-07-06 CJ: added IamIPv6Ready()
 *   2005-05-30 CJ: added prefix longest matching support
 *                  added addr_scope()
 *   2005-05-27 CJ: added gatewaying support
 *   2005-01-26 cft: reformatted and integrated into Lunar trunk
 *
 * This file is distributed under the terms of the GNU general Public
 * License (GPL), see the file COPYING
 */

#ifdef LUNAR_HAS_IPV6

#define ICMP6_BASE_LEN 8

// These (below) are really checked by the IPv6 stack in the kernel

#define IN6ADDR_ALL_NODES { { { 0xff,0x02,0,0,0,0,0,0,0,0,0,0,0,0,0,1 } } } 
#define IN6ADDR_FAKE_SRC { { { 0xfe,0x80,0,0,0,0,0,0,0x02,0x11,0x22,0xff,0xfe,0x33,0x44,0x55 } } }
#define ETH_FAKE_SRC { 0x00,0x11,0x22,0x33,0x44,0x55 }

int remaining_len = 0;

/*
static uint16_t check( int count, uint16_t *data )
{
	uint32_t sum = 0;
	
	while( count > 1 ) {
	
		sum += *(data++);
		count -= 2;
	}
	
	if( count > 0 )
		sum += *(unsigned char *)data;
		
	while( sum>>16 )
		sum = (sum & 0xffff) + (sum>>16);
		
	return (~sum);
}*/

char addr_scope(struct in6_addr *addr)
{
	if (IN6_IS_ADDR_LINKLOCAL(addr))
		return IPV6_ADDR_SCOPE_LINKLOCAL;
		
	if (IN6_IS_ADDR_SITELOCAL(addr))
		return IPV6_ADDR_SCOPE_SITELOCAL;
	
	if (IN6_IS_ADDR_MANETLOCAL(addr))
		return IPV6_ADDR_SCOPE_MANET;
	
	return IPV6_ADDR_SCOPE_GLOBAL;
}

int prefix_match_len(void *addr1, void *addr2)
{
   /*
   *  find the first different bit between two addresses
   *  length of address must be a multiple of 32bits
   *
   *  Slightly modified copy of net/ipv6/ip6_fib.c from linux kernel source code
   *  --> the value returned is the length (in bits) of the common part
   */
	
	__u32 *a1 = addr1;
	__u32 *a2 = addr2;
	int i;
	
	for (i = 0; i < 4; i++) {
		__u32 xb;

		xb = a1[i] ^ a2[i];

		if (xb) {
			int j = 31;

			xb = ntohl(xb);

			while ((xb & (1 << j)) == 0)
				j--;

			return (i * 32 + 31 - j);
		}
	}
	
	return 128;	// addresses are identical !
}

int IamIPv6Ready(void)
{
	struct inet6_dev *in6dev = (struct inet6_dev*) lunardev->ip6_ptr;
	struct inet6_ifaddr *cur;
			
	if (in6dev && in6dev->addr_list) {
		for (cur = in6dev->addr_list; cur; cur = cur->if_next) {
			if (!IN6_IS_ADDR_UNSPECIFIED(&cur->addr))
				return 1;
		}
	}
	
	return 0;
}


void lunar_ipv6_send_RA(struct in6_addr *ipv6_prefix, struct in6_addr *rt_addr, int rt_lifetime, unsigned char addr_lifetime )
{
	unsigned char buf[512];
	struct ipv6hdr *ip6h = (struct ipv6hdr*) buf;
	struct ra_msg *ra = (struct ra_msg *) (buf + sizeof(struct ipv6hdr));
	unsigned char *opt = (char*) (buf + sizeof(struct ipv6hdr) + sizeof(struct ra_msg));
	int len;
	unsigned char pseudo_hdr[128]; // for checksum calculation
	
	struct in6_addr all_ipv6_nodes = IN6ADDR_ALL_NODES;
	struct in6_addr ipv6_fake_src = IN6ADDR_FAKE_SRC;
	// char eth_fake_src[ETH_ALEN] = ETH_FAKE_SRC;
	
	memset( buf, 0, 512 );
	
	// IPv6 header
	*buf = 0x60; // version
	ip6h->hop_limit = 255;
	ip6h->nexthdr = NEXTHDR_ICMP;
	ip6h->payload_len = htons( (sizeof(struct ra_msg) +32) ); // see options
	
	// IPv6 DST address 
	memcpy( &(ip6h->daddr), &all_ipv6_nodes, 16 );
	
	// fake SRC address : just for LUNAR-MANET-local
	memcpy( &(ip6h->saddr), &ipv6_fake_src, sizeof(struct in6_addr));
#ifdef LUNAR_HAS_GW
	// default route is created with link-local src addr of IPv6 header	
	if(rt_addr)
		memcpy( &(ip6h->saddr), rt_addr, sizeof(struct in6_addr));	
#endif
	
	// RA header (including icmp6 header)
	// Note that the router lifetime is set to zero
	// by default (for LUNAR-MANET-local)
	// so that router is not a default route
#ifdef LUNAR_HAS_GW	
	if(rt_lifetime)
		ra->icmph.icmp6_rt_lifetime = htons(rt_lifetime);
#endif
	ra->icmph.icmp6_type = NDISC_ROUTER_ADVERTISEMENT;
	ra->icmph.icmp6_code = 0;
	ra->reachable_time = htonl(0);
	ra->retrans_timer = htonl(0);
	
	/* CJ: because with gatewaying we cannot give the real
	   eth address of the gateway, we skip the eth source
	   option --> allowed by RFC 2461
		
	// ETH SRC OPTION : 8 bytes
	*opt = ND_OPT_SOURCE_LL_ADDR;
	*(opt+1) = 1; // length of option
	memcpy( opt+2, eth_fake_src, 6 );
	
	opt += 8;  // offset in option list*/
	
	// PREFIX OPTION : 32 bytes
	
	*opt = ND_OPT_PREFIX_INFO;
	*(opt+1) = 4; // length of option
	*(opt+2) = 64; // prefix length;
	*(opt+3) = 0xc0; // 0x40 for off-link address, 0xc0 for on-link 
	memset( opt+4, addr_lifetime, 8 );  // addr lifetime 0 or infinity
	memset( opt+12, 0, 4 );  // reserved NULL field
	memcpy( opt+16, ipv6_prefix, sizeof(struct in6_addr)); // copy prefix 
	
	// build PSEUDO HEADER for checksum calculation
	
	memset( pseudo_hdr, 0, 128 );
	memcpy( pseudo_hdr, &(ip6h->saddr), 16 );
	memcpy( pseudo_hdr+16, &(ip6h->daddr), 16 );
	memcpy( pseudo_hdr+32, &(ip6h->payload_len), 2 );
	memcpy( pseudo_hdr+39, &(ip6h->nexthdr), 1 );
	memcpy( pseudo_hdr+40, ra, ntohs(ip6h->payload_len) );
	
	len = ntohs(ip6h->payload_len) + sizeof(struct ipv6hdr);
	ra->icmph.icmp6_cksum = check( len, (u_int16_t*)pseudo_hdr );

	netbox_l3deliver(lunardev->ifindex, buf, len);
}

int parse_icmp6(sapfpkt *pkt)
{
	struct icmp6hdr *icmp6h = (struct icmp6hdr *)(pkt->data+ETH_HLEN+sizeof(struct ipv6hdr));
	struct nd_msg *ndh = (struct nd_msg *)(pkt->data+ETH_HLEN+sizeof(struct ipv6hdr));
	
	struct lunartarget_s target;
	sapfsel_t sel = 0;
	
	DEBUG(9, printk("ICMP6 type : %d\n", icmp6h->icmp6_type ));
	DEBUG(9, printk("ICMP6 code : %d\n", icmp6h->icmp6_code ));
	DEBUG(9, printk("ICMP6 chksum : %d\n", ntohs(icmp6h->icmp6_cksum) ));

	// remove ICMP6 header size from payload len
	remaining_len -= 8;

	switch(icmp6h->icmp6_type)
	{
		case NDISC_ROUTER_SOLICITATION: {
			struct in6_addr lunar_ipv6_prefix = LUNAR_IPV6_PREFIX;
			DEBUG(9, printk("ICMP6 --> Router Solicitation\n" ));
			
			// the only option can be ETH SRC address
			if( remaining_len == 8 ) 
			{
				DEBUG(9, printk("SRC ETH : %s\n",
						lunar_eth2str(pkt->data+ETH_HLEN+sizeof(struct ipv6hdr)+ICMP6_BASE_LEN+2)));
			}
			// we send a fake RA to the IPv6 stack
			lunar_ipv6_send_RA(&lunar_ipv6_prefix, 0, 0, 0xff);
			break;
		}
		case NDISC_ROUTER_ADVERTISEMENT:
			DEBUG(9, printk("ICMP6 --> Router Advertisement\n" ));
			break;
		case NDISC_NEIGHBOUR_SOLICITATION:
			DEBUG(9, printk("ICMP6 --> ND Solicitation\n" ));
			DEBUG(9, printk("TARGET : %s\n",
					lunar_ip62str(&(ndh->target))));
			remaining_len -= sizeof(struct in6_addr);
			// the only remaining option can be ETH SRC address
			if( remaining_len == 8 ) {
				DEBUG(9, printk("SRC ETH : %s\n",
					     lunar_eth2str((ndh->opt)+2 )));
			}
			
			if( IN6_IS_ADDR_LINKLOCAL(&ndh->target) ) {
				// do nothing for the moment
			} else {
				memcpy(&(target.id.ipv6),
				       &ndh->target, 16);
				target.family = AF_INET6;
				lunar_do_ndisc(&target, 0, 0, AF_INET6);
			}
			break;
		case NDISC_NEIGHBOUR_ADVERTISEMENT:
			DEBUG(9, printk("ICMP6 --> ND Advertisement\n" ));
			break;
		case ICMPV6_ECHO_REQUEST:
			DEBUG(9, printk("ICMP6 --> ECHO request : sending ! " ));
			memcpy(((char*)&sel) + sizeof(sel) - ETH_ALEN, pkt->data,ETH_ALEN);
			DEBUG(9, printk("(sel : 0x%016Lx)\n", sel ));
			
			pkt->len = pkt->len - ETH_HLEN;
			pkt->data = pkt->data + ETH_HLEN;
			sapf_demux(mysapfhandle, sel, pkt);
			break;
		case ICMPV6_ECHO_REPLY:
			DEBUG(9, printk("ICMP6 --> ECHO reply : sending ! " ));
			memcpy(((char*)&sel) + sizeof(sel) - ETH_ALEN, pkt->data,ETH_ALEN);
			DEBUG(9, printk("(sel : 0x%016Lx)\n", sel ));
			
			pkt->len = pkt->len - ETH_HLEN;
			pkt->data = pkt->data + ETH_HLEN;
			sapf_demux(mysapfhandle, sel, pkt);
			break;
		default:
			DEBUG(9, printk("ICMP6 --> Unsupported Message\n" ));
			break;
			
	}

	return 1;
}

int parse_v6_data(sapfpkt *pkt)
{
#ifndef NO_DEBUG
	struct ethhdr *eth_hdr = (struct ethhdr *)(pkt->data);
#endif
	struct ipv6hdr *ip6h = (struct ipv6hdr *)(pkt->data+ETH_HLEN);
	
	DEBUG(9, printk("Parsing (IPv6) data ...\n"));
	
	DEBUG(9, printk("ETH SRC : %s\n", lunar_eth2str(eth_hdr->h_source)));
	DEBUG(9, printk("ETH DST : %s\n", lunar_eth2str(eth_hdr->h_dest)));
		
	DEBUG(9, printk("IPv6 version : %02x\n", *(char*)ip6h ));
	DEBUG(9, printk("IPv6 payload length : %d bytes\n", ntohs(ip6h->payload_len) ));
	DEBUG(9, printk("IPv6 hop limit : %d\n", ip6h->hop_limit ));
	DEBUG(9, printk("IPv6 SRC : %s\n", lunar_ip62str(&ip6h->saddr)));
	DEBUG(9, printk("IPv6 DST : %s\n", lunar_ip62str(&ip6h->daddr)));
	DEBUG(9, printk("IPv6 next header : %d ", ip6h->nexthdr ));
	
	remaining_len = ntohs(ip6h->payload_len);
	
	switch( ip6h->nexthdr )
	{
		case NEXTHDR_HOP:
			DEBUG(9, printk("(HOP-by-HOP Option)\n"));
			break;
		case NEXTHDR_ICMP:
			DEBUG(9, printk("(ICMP6)\n"));
			parse_icmp6(pkt);
			break;
		case NEXTHDR_TCP:
			DEBUG(9, printk("(TCP)\n"));
			break;
		case NEXTHDR_UDP:
			DEBUG(9, printk("(UDP)\n"));
			break;
		default:
			DEBUG(9, printk("(Unsupported header)\n"));
			break;
	}

	return 1;
}

#endif // LUNAR_HAS_IPV6

// eof
