/* sapf.c
 *
 *   SAPF kernel module
 *
 *   A partial implementation of the Simple Active Packet Format
 *
 * Copyright (C) 2001-2004 <christian.tschudin@unibas.ch>
 *
 * This file is distributed under the terms of the GNU general Public
 * License (GPL), see the file COPYING
 *
 *   2004-11-13 Added support for kernel 2.6
 *              (Per Svensson)
 *   2003-11-23 removed linux kernel code for portability (user space
 *              as well as windows)
 *   2003-11-02 some bug fixes
 *   2003-05-10 added support for multiple named fib tables
 *   2002-12-24 selectors now stored in network order, removed two '&&' bugs
 *   2002-12-21 permit NULL callback ie black hole (or use a redirect to 0)
 *   2002-12-03 added activity flag, export the new testandreset symbol
 *   2002-10-11 added multi-client support (explicit open, returns handle)
 *   2002-09-17 fixed infinite loop in hash table fine, discovered by
                David Rosen and Olof Rensfelt
 *   2002-06-20 kernelized for Linux
 *   2001-11-18 initial version for SelNet
 */


#include "../include/sapf.h"
#include "../include/sapf_decl.h"

static int sapfdebug = 0;

#ifndef NS_PORT
  static struct timer_list sapf_gc_timer;
#endif


// ----------------------------------------------------------------------

 int SAPF_CLASS sel2hash(int handle, sapfsel_t s)
{
	return (int) ((unsigned int)(s ^ ((s & htonsel(SELMASK)) >> 32)) % fib[handle & 0x0f]->fibsize);
}

 int SAPF_CLASS sel2ndx(int handle, sapfsel_t s) // return ndx of matching entry or -1
{
	int i = sel2hash(handle, s), j;
	int fibsize = fib[handle & 0x0f]->fibsize;
	struct fibentry_s *f =  fib[handle & 0x0f]->tab + i;

	for (j = fibsize; j > 0; j--) {
		if ((f->flags_and_op & FE_MASK) == FE_EMPTY)
			return -1;
		if ((f->flags_and_op & FE_MASK) == FE_USED && f->sel == s)
			return i;
		// skip over deleted or missmatching entries:
		if (++i >= fibsize) {
			i = 0;
			f = fib[handle & 0x0f]->tab;
		} else
			f++;
	}

	return -1;
}

 int SAPF_CLASS sel2EmptyNdx(int handle, sapfsel_t s)
{
	int i = sel2hash(handle, s), cnt = 0;
	int fibsize = fib[handle & 0x0f]->fibsize;
	struct fibentry_s *f =  fib[handle & 0x0f]->tab + i;

	while ((f->flags_and_op & FE_MASK) == FE_USED) {
		if (++cnt >= fibsize) {
			SAPFDEBUG(1,printk("sapf: running out of free SAPF slots\n"));
			return -1;
		}
		if (++i >= fibsize) {
			i = 0;
			f = fib[handle & 0x0f]->tab + i;
		} else
			f++;
	}
	memset(f, 0, sizeof(struct fibentry_s));
	return i;
}

 int SAPF_CLASS freeEntry(int handle, int i) // mark this entry as deleted, release aux memory
{
	struct fibentry_s *f = fib[handle & 0x0f]->tab + i;
	SAPFDEBUG(8, printk("sapf - removing entry %016Lx\n", ntohsel(f->sel)));

	f->flags_and_op = (f->flags_and_op & ~FE_MASK) | FE_DELETED;
	if ((f->flags_and_op & SAPFOP_MASK) == SAPFOP_CBACK &&
	    		f->p2.aux && f->credit != -1) {
		kfree(f->p2.aux);
		f->p2.aux = 0;
	}

	return 0;
}

 void SAPF_CLASS selRandomize(sapfsel_t *sel)
{
	get_random_bytes(sel, sizeof(sapfsel_t));
}

// ----------------------------------------------------------------------

int SAPF_CLASS sapf_init(void)
{
        SAPFDEBUG(2, printk("sapf_init\n"));

#ifndef NS_PORT
 	init_timer(&sapf_gc_timer);
	sapf_gc_timer.function = sapf_gc;
#endif //NS_PORT
	//Common to both real and NS_PORT
        arm_timer(&sapf_gc_timer, jiffies + HZ);

	return 0;

}

int SAPF_CLASS sapf_open(char *name, int size)
{
	int i;
	int t = -1;

	SAPFDEBUG(2, printk("sapf_open \"%s\" (%d)\n", name, size));

	for (i = 0; i < SAPF_MAX_FIB_TABLES; i++) {
		if (!fib[i] || !fib[i]->fibsize) {
 			if (t == -1)
				t = i;
			continue;
		}

		if (!strcmp(name, fib[i]->name))
			break;
	}

 	if (i >= SAPF_MAX_FIB_TABLES) {
		int len;


		if (t == -1) {
			SAPFDEBUG(1,printk("sapf_open - too many FIBs\n"));
			return -1;
		}

		len = sizeof(struct fibtable_s) + size * sizeof(struct fibentry_s);
		fib[t] = (struct fibtable_s*) kmalloc(len, GFP_KERNEL);
		if (!fib[t]) {
			SAPFDEBUG(1,printk("sapf - no memory for fib!\n"));
			return -1;
		}
		memset(fib[t], 0, len);
		fib[t]->fibsize = size;
		selRandomize(&(fib[t]->nextselval));
		strncpy(fib[t]->name, name, sizeof(fib[i]->name) - 1);
	} else
		t = i;

	fib[t]->opencnt++;

	for (i = 0; i < SAPF_MAX_CLIENTS; i++) {
		if (!clients[i]) {
			SAPFDEBUG(10, printk("sapf_open / client is %d, handle is %x\n", i, (i<<8)|t));
			clients[i] = 1;
//			UP_MODULE_USE();
			return (i << 8) | t;
		}
	}
	/// FIXME: we should also undo the opencnt++

	SAPFDEBUG(10, printk("sapf_open ended with -1\n"));
	return -1;
}

void SAPF_CLASS sapf_close(int handle)
{
	struct fibentry_s *f;
	int i, cl = handle >> 8;
	struct fibtable_s *ftab;

	SAPFDEBUG(2, printk("sapf_close 0x%x\n", handle));

	if (handle == -1)
		return;
	if (!clients[cl]) {
		SAPFDEBUG(1,printk("sapf_close: empty client %d\n", cl));
		return;
	}

	ftab = fib[handle & 0x0f];

	for (i = 0, f = ftab->tab; i < ftab->fibsize; i++, f++) {
		if ((f->flags_and_op & FE_MASK) == FE_USED &&
		    f->client == cl)
			freeEntry(handle, i);
	}
	ftab->opencnt--;
	if (ftab->opencnt == 0) {
		fib[handle & 0x0f] = 0;
		kfree(ftab);
	}

	clients[cl] = 0;
//	DOWN_MODULE_USE();
}

int SAPF_CLASS sapf_demux(int handle, sapfsel_t sel, sapfpkt *pkt)
{
	struct fibentry_s *f;
	int ndx;

	for (;;) {
	  SAPFDEBUG(9, printk("sapf - demuxing selector 0x%016Lx\n", ntohsel(sel)));
		if (sel == 0) {
			SAPFDEBUG(9, printk("  NULL selector\n"));
			return 0;
		}
		ndx = sel2ndx(handle, sel);
		if (ndx < 0) {
			SAPFDEBUG(9, printk("  not found\n"));
			return -1;
		}

		f = fib[handle & 0x0f]->tab + ndx;
		if (f->credit >= 0 ) {
			if (f->credit < FIB_DECAY_PER_PKT) {
				SAPFDEBUG(9, printk("  expired\n"));
				freeEntry(handle, ndx);
				return -1;
			} else {
				f->credit -= FIB_DECAY_PER_PKT;
				f->flags_and_op |= FE_ACTIVE;
			}
		}

		switch (f->flags_and_op & SAPFOP_MASK) {
		case SAPFOP_NAT:
		case SAPFOP_CBACK:
			SAPFDEBUG(9, printk("  callback\n"));
			if (NS_CLASS_Instance_PTR f->p1.callback)
				(NS_CLASS_Instance_PTR f->p1.callback)(pkt, f->p2.aux);
			return 0;
		case SAPFOP_REDIR:
			SAPFDEBUG(9, printk("  redir -> 0x%016Lx\n", f->p1.sel));
			sel = f->p1.sel;
			break;
		default:
			SAPFDEBUG(1, printk("  unknown opcode 0x%x\n",
			    ((f->flags_and_op & SAPFOP_MASK) >> SAPFOP_SHIFT)));
			return -1;
		}
	}
}

 void SAPF_CLASS sapf_gc(unsigned long dummy)
{
	struct fibentry_s *f;
	int i, t;

        SAPFDEBUG(10, printk("sapf_gc\n"));

	for (t = 0; t < SAPF_MAX_FIB_TABLES; t++) {
		if (!fib[t])
			continue;
		for (i = 0, f = fib[t]->tab; i < fib[t]->fibsize; i++, f++) {
			if ((f->flags_and_op & FE_MASK) != FE_USED || f->credit < 0)
				continue;
			if (f->credit < FIB_DECAY_PER_SEC) {
				freeEntry(t, i);
			} else
				f->credit -= FIB_DECAY_PER_SEC;
		}
	}

        arm_timer(&sapf_gc_timer, jiffies + HZ);

}

void SAPF_CLASS sapf_fini(void)
{
	int i, t;

        SAPFDEBUG(2, printk("sapf_fini\n"));

        cancel_timer(&sapf_gc_timer);

	for (t = 0; t < SAPF_MAX_FIB_TABLES; t++) {
		struct fibentry_s *f;

		if (!fib[t])
			continue;
		for (i = 0, f = fib[t]->tab; i < fib[t]->fibsize; i++, f++) {
			if ((f->flags_and_op & FE_MASK) == FE_USED)
				freeEntry(t, i);
		}
		kfree(fib[t]);
		fib[t] = 0;
	}
}

// ----------------------------------------------------------------------

int SAPF_CLASS sapf_exists(int handle, sapfsel_t sel)
{
	return sel2ndx(handle, sel) >= 0;
}

sapfsel_t SAPF_CLASS sapf_getUnusedSelector(int handle, sapfsel_t class_)
{
	sapfsel_t s;
	struct fibtable_s *ftab = fib[handle & 0x0f];
	int i;

	if (!ftab) {
		SAPFDEBUG(1,printk("ksapf: ouch, wrong handle 0x%0x\n", handle));
		return 0;
	}

	for (i = 100; i > 0; i--) {
		s = ftab->nextselval & SELMASK & ~SELCLASS_MASK;

		//	nextselval *= nextselval;
		//	nextselval++;  
#ifndef NS_PORT
		ftab->nextselval *=11;
#else // NS_PORT is defined
		ftab->nextselval ++;
#endif
		if (ftab->nextselval == 0 || ftab->nextselval == 1)
			selRandomize(&(ftab->nextselval));
		s = htonsel(s | class_);
		if (sel2ndx(handle, s) < 0)
			return s;
	}
	SAPFDEBUG(1,printk("ksapf: out of unused selectors for handle 0x%0x\n", handle));
	return 0;
}

int SAPF_CLASS sapf_addCallback(int handle, sapfsel_t sel, sapfcallback_t fct, void *aux)
{
	struct fibentry_s *f;
	int ndx;

	SAPFDEBUG(8, printk("sapf - adding callback for %016Lx\n", ntohsel(sel)));

	ndx = sel2ndx(handle, sel);
	if (ndx < 0) {
		ndx = sel2EmptyNdx(handle, sel);
		if (ndx < 0)
			return -1;
	}
	f = fib[handle & 0x0f]->tab + ndx;
	f->sel = sel;
	f->client = handle >> 8;
	f->credit = FIB_START_CREDIT;
	f->p1.callback = fct;
	f->p2.aux = aux;

	if( fct == netbox_nat )
		f->flags_and_op = FE_USED | SAPFOP_NAT;
	else
		f->flags_and_op = FE_USED | SAPFOP_CBACK;

	return 0;
}

void* SAPF_CLASS sapf_getCallbackAux(int handle, sapfsel_t sel)
{
	struct fibentry_s *f;
	int ndx;

	SAPFDEBUG(8, printk("sapf - getting callback aux-ptr for %016Lx\n", ntohsel(sel)));

	ndx = sel2ndx(handle, sel);
	if (ndx < 0)
		return 0;

	f = fib[handle & 0x0f]->tab + ndx;
	return (f->flags_and_op & SAPFOP_MASK) == SAPFOP_CBACK ? f->p2.aux : 0;
}

int SAPF_CLASS sapf_setPermanent(int handle, sapfsel_t sel)
{
	int ndx = sel2ndx(handle, sel);

	if (ndx < 0)
		return -1;
	fib[handle & 0x0f]->tab[ndx].credit = -1;
	return 0;
}

int SAPF_CLASS sapf_deleteEntry(int handle, sapfsel_t sel) // remove entry, if existing
{
	int ndx = sel2ndx(handle, sel);

	if (ndx < 0)
		return -1;
	freeEntry(handle, ndx);
	return 0;
}

int SAPF_CLASS sapf_addREDIR(int handle, sapfsel_t sel, sapfsel_t next)
{
	struct fibentry_s *f;
	int ndx;

	ndx = sel2ndx(handle, sel);
	if (ndx < 0) {
		ndx = sel2EmptyNdx(handle, sel);
		if (ndx < 0)
			return -1;
	}
	f = fib[handle & 0x0f]->tab + ndx;
	f->sel = sel;
	f->client = handle >> 8;
	f->flags_and_op = FE_USED | SAPFOP_REDIR;
	f->credit = FIB_START_CREDIT;
	f->p1.sel = next;
	f->p2.aux = 0;
    
	return 0;
}


int SAPF_CLASS sapf_testandreset_active(int handle, sapfsel_t sel)
{
	struct fibentry_s *f;
	int ndx;
	int rc;

	ndx = sel2ndx(handle, sel);
	if (ndx < 0)
		return 0;

	f = fib[handle & 0x0f]->tab + ndx;
	rc = (f->flags_and_op & FE_ACTIVE) == FE_ACTIVE;
	f->flags_and_op &= ~FE_ACTIVE;

	return rc;
}

// eof
